from dataclasses import field
from typing import List
from marshmallow import validate
from marshmallow_dataclass import dataclass, Optional

valid_as_ccodes = ("j", "e", "o", "h", "J", "C", "g", "G", "=", "n", "_")
redundant_as_ccodes = ("j", "n", "O", "e", "o", "h", "J", "C", "c", "m", "mo", "=", "_", "x", "p", "P", "X", "I", "i")


@dataclass
class AlternativeSplicingConfiguration:
    report: bool = field(default=True, metadata={
        "metadata": {
            "description": "Boolean flag, about whether Mikado should find and report AS events \
or simply define one transcript per locus."}
    })
    cds_only: bool = field(default=False, metadata={
        "metadata": {
            "description": "Only consider the coding part of transcripts to determine the validity of the AS event."},
    })
    min_cds_overlap: float = field(default=0.5, metadata={
        "metadata": {
            "description": "Minimum CDS overlap threshold (for coding transcripts) to be cleared for two transcripts to be considered AS events of each other.",
            }, "validate": validate.Range(min=0, max=1)
    })
    min_cdna_overlap: float = field(default=0.6, metadata={
        "metadata": {
            "description": "Minimum cDNA overlap threshold to be cleared for two transcripts to be considered AS events of each other.",
            }, "validate": validate.Range(min=0, max=1)
    })
    keep_retained_introns: bool = field(default=True, metadata={
        "metadata": {"description": "Keep or discard AS events with a retained intron. Default: true"},
    })
    keep_cds_disrupted_by_ri: bool = field(default=False, metadata={
        "metadata": {
            "description": "Keep or discard AS events with their CDS disrupted by a retained intron event, ie either having their stop codon or ending with a truncated CDS within the intron of another transcript. Default: false."},
    })
    max_isoforms: int = field(default=10, metadata={
        "metadata": {"description": "Maximum number of isoforms to report per locus. Default: 10.",},
        "validate": validate.Range(min=1)
    })
    # Note: valid and redundant class codes will be changed TEMPORARILY AT RUN TIME, FOR EACH LOCUS CLASS,
    # if the padding is enabled. Briefly, locus classes will need to *initially* accept into the locus transcripts that
    # would normally be considered redundant (e.g. '=' or 'n') as they could act as templates. Note that normally these
    # additional class codes would *not* be valid choices as AS events.
    # AFTER the padding is completed, these extra class-codes will be removed from the valid class codes list and
    # re-added to the redundant class code list, therefore enabling the removal of redundancy from the final locus.
    # This happens during the __init__ of Locus in Mikado/loci/locus.py (adding of valid class codes) and later they are
    # removed before the finalisation. Note that these modified codes are a *copy* of the codes of the configuration.
    valid_ccodes: list = field(default_factory=lambda: ["j", "J", "G", "h"], metadata={
        "metadata": {
            "description": "AS event class codes considered as valid AS events. Valid codes are in categories \
'Alternative splicing', 'Extension' (with junction F1 lower than 100%), and 'Overlap' (exluding m). \
Please run 'mikado util class_codes' or refer to the online documentation for an explanation of each code.",},
        "validate": validate.ContainsOnly(list(valid_as_ccodes))
    })
    redundant_ccodes: list = field(default_factory=lambda: ["c", "m", "_", "=", "n"], metadata={
        "metadata": {
            "description": "AS event class codes considered as a duplicate of other transcripts in the locus. \
Please run 'mikado util class_codes' or refer to the online documentation for an explanation of each code."},
        "validate": validate.ContainsOnly(list(redundant_as_ccodes))
    })
    min_score_perc: float = field(default=0.5, metadata={
        "metadata": {
            "description": "Minimum percentage of the score associated to an AS event *compared to the primary \
transcript* for the AS event to be considered valid. Default: 0.5, or 50%.",},
            "validate": validate.Range(min=0, max=1)
    })
    only_confirmed_introns: bool = field(default=True, metadata={
        "metadata": {"description": "Boolean flag. If set to true (default), Mikado will only report AS \
events whose introns *not in common with the primary transcript* are verified by the junctions provided to \
serialise (usually Portcullis reliable junctions)."},
    })
    ts_distance: int = field(default=2000, metadata={
        "metadata": {
            "description": "When padding, this value indicates how many bps can be added to *either* of the 5' or 3' \
section of the transcript, excluding introns.",},
        "validate": validate.Range(min=0)
    })
    pad: bool = field(default=True, metadata={
        "metadata": {"description": "Boolean flag. If set to true, Mikado will pad transcripts. \
Please refer to the online documentation."},
        "required": True
    })
    ts_max_splices: int = field(default=2, metadata={
        "metadata": {
            "description": "When padding, this value indicates the maximum number of splicing junctions \
that can be added to *either* of the 5' or 3' section of the transcript."},
            "validate": validate.Range(min=0)
    })


@dataclass
class OutputFormatConfiguration:
    source: str = field(default="Mikado", metadata={
        "metadata": {"description": "Prefix for the source field in the mikado output."},
    })
    id_prefix: str = field(default="mikado", metadata={
        "metadata": {"description": "Prefix for the ID of the genes/transcripts in the output"},
    })
    report_all_orfs: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean switch. If set to true, Mikado will report all ORFs associated with a \
transcript in the final loci file."},
    })
    report_all_external_metrics: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean switch. If set to True, Mikado will report for each transcript all available \
external metrics, not just those requested for in the scoring file. On databases with many external scores \
(e.g. in Minos), this could negatively affect performance."
        }
    })


@dataclass
class OrfLoadingConfiguration:
    minimal_secondary_orf_length: int = field(default=200, metadata={
        "metadata": {
            "description": "Minimum length of a *secondary* ORF to be loaded after the first, in bp. Default: 200 bps"},
        "validate": validate.Range(min=0)
    })
    minimal_orf_length: int = field(default=50, metadata={
        "metadata": {
            "description": "Minimum length in bps of an ORF to be loaded, as the primary ORF, onto a transcript. \
Default: 50 bps"},
        "validate": validate.Range(min=0)
    })
    strand_specific: bool = field(default=True, metadata={
        "metadata": {
            "description": "Boolean flag. If set to true, monoexonic transcripts with an available ORF \
on the opposite strand will still not be reversed."},
    })


@dataclass
class BlastParamsConfiguration:
    evalue: float = field(default=1e-06, metadata={
        "metadata": {"description": "Minimum evalue for the whole hit. Default: 1e-6"},
        "validate": validate.Range(min=0)
    })
    hsp_evalue: float = field(default=1e-06, metadata={
        "metadata": {
            "description": "Minimum evalue for any HSP hit (some might be discarded even if the whole hit is valid). Default: 1e-6"},
        "validate": validate.Range(min=0)
    })
    leniency: str = field(default="STRINGENT", metadata={
        "metadata": {
            "description": "One of 'STRINGENT', 'LENIENT', 'PERMISSIVE'. Please refer to the online documentation \
for details. Default: STRINGENT"},
        "validate": validate.OneOf(["STRINGENT", "LENIENT", "PERMISSIVE"]),
        "required": True
    })
    max_target_seqs: int = field(default=3, metadata={
        "metadata": {"description": "Maximum number of hits to consider. Default: 3"},
        "validate": validate.Range(min=1)
    })
    minimal_hsp_overlap: float = field(default=0.5, metadata={
        "metadata": {
            "description": "Minimum overlap of the ORF with the HSP (*not* reciprocal). Default: 0.8, i.e. 80%"},
        "validate": validate.Range(min=0, max=1)
    })
    min_overlap_duplication: float = field(default=0.8, metadata={
        "metadata": {
            "description": "min_overlap_duplication: minimum overlap (in %) for two ORFs to consider them as \
target duplications. This means that if two ORFs have no HSPs in common, but the coverage of their disjoint \
HSPs covers more than this percentage of the length of the *target*, they represent most probably a duplicated gene."},
        "validate": validate.Range(min=0, max=1)
    })


@dataclass
class ChimeraSplitConfiguration:
    blast_check: bool = field(default=True, metadata={
        "metadata": {
            "description": "Whether to use BLAST information to take a decision. See blast_params for details."},
        "required": True
    })
    execute: bool = field(default=True, metadata={
        "metadata": {"description": "Whether to split multi-ORF transcripts at all. Boolean."},
        "required": True
    })
    skip: List[bool] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "Input sources for which Mikado will skip the splitting, e.g. ultra-reliable \
full cDNA sequences."},
        "validate": validate.Length(min=0)
    })
    blast_params: BlastParamsConfiguration = field(default_factory=BlastParamsConfiguration, metadata={
                "metadata": {"description": "Parameters for the BLAST check prior to splitting."},
    })


@dataclass
class RunOptionsConfiguration:
    # From swagger-marshmallow-codegen (Unique validator)
    class Unique(validate.Validator):
        message = "{input} is Not unique"

        def __init__(self, error=None):
            self.error = error

        def _repr_args(self):
            return ""

        def _format_error(self, value):
            return self.message.format(input=value)

        def __call__(self, value):
            if len(value) != len(set(value)):
                raise validate.ValidationError(self._format_error(value))
            return value

    class MinLength(validate.Validator):
        message = "{input} contains values below 1"
        def __init__(self, error=None):
            self.error = error

        def _repr_args(self):
            return ""

        def _format_error(self, value):
            return self.message.format(input=value)

        def __call__(self, value):
            if min(value) < 1:
                raise validate.ValidationError(self._format_error(value))
            return value

    shm: bool = field(default=False, metadata={
        "metadata": {
            "description": "boolean flag. If set and the DB is sqlite, it will be copied onto the /dev/shm faux \
partition, for a potentially faster execution."},
    })
    exclude_cds: bool = field(default=False, metadata={
        "metadata": {
            "description": "boolean flag. If set, the CDS information will not be printed in Mikado output. \
Default: false"},
    })
    intron_range: List[int] = field(default_factory=lambda: [60, 10000], metadata={
        "metadata": {
            "description": "A range where most of the introns (99%) should fall into. Transcripts with too many \
introns larger or smaller than what is defined in this range will be penalised in the scoring. Default: [60, 900]"},
        "validate": [validate.Length(min=2, max=2), Unique, MinLength],
        "required": True
    })
    reference_update: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set, Mikado will run in reference-update mode, see documentation."},
    })
    only_reference_update: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set, Mikado will run in reference-update mode, see documentation. \
Additionally, Mikado will ignore any locus where there is not at least one reference transcript."},
    })
    check_references: bool = field(default=False, metadata={
        "metadata": {
            "description": "boolean flag. If set to true, transcripts marked as reference will still \
be checked for compliance with the requirements established in the scoring file."},
    })
    single_thread: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set, multithreading will be disabled - useful for profiling and debugging."},
    })


@dataclass
class ClusteringConfiguration:
    cds_only: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean, it specifies whether to cluster transcripts only according to their CDS (if "
                           "present). Please note that this applies *only* when comparing pairs of coding "
                           "transcripts. If *either* transcript under consideration is non-coding, Mikado will "
                           "consider both coding and non-coding parts of the transcript. for assessing the "
                           "clustering."},
    })
    min_cds_overlap: float = field(default=0.2, metadata={
        "metadata": {
            "description": "Minimal CDS overlap for the second clustering, in percentage between 0 and 1. \
Default: 0.2, or 20%"
        },
        "validate": validate.Range(min=0, max=1)
    })
    min_cdna_overlap: float = field(default=0.2, metadata={
        "metadata": {
            "description": "Minimal cDNA overlap for the second clustering, in percentage between 0 and 1. \
Default: 0.2, or 20%."
        },
        "validate": validate.Range(min=0, max=1)
    })
    purge: bool = field(default=True, metadata={
        "metadata": {
            "description": "Boolean, it specifies whether to remove transcripts which fail the minimum requirements \
check, or if instead to just assign them a score of 0 (potentially retaining them in the final output)."},
    })
    flank: int = field(default=200, metadata={
        "metadata": {"description": "Maximum distance for transcripts to be clustered within the same superlocus."},
        "validate": validate.Range(min=0)
    })
    simple_overlap_for_monoexonic: bool = field(default=False, metadata={
        "metadata": {
            "description": "boolean. Disabled by default. If set to true, then any overlap, even minimal, \
will suffice to incude monoexonic transcripts in a locus."},
    })


@dataclass
class FragmentsConfiguration:
    remove: bool = field(default=True, metadata={
        "metadata": {
            "description": "boolean. Whether to remove fragments or leave them, properly tagged, in the output file. \
Default: remove them."}
    })
    max_distance: int = field(default=2000, metadata={
        "metadata": {"description": "Maximum distance of a putative fragment from a valid gene, for it \
to be considered by this filter."},
            "validate": validate.Range(min=0)
    })
    valid_class_codes: List[str] = field(
        default_factory=lambda: ["p", "P", "x", "X", "i", "m", "_", "e", "o"], metadata={
            "metadata": {
                "description": "Which class codes will be considered as fragments. Default: (p, P, x, X, i, m, _). \
Choices: '_' plus any class code with category 'Intronic', 'Fragment', or 'Overlap'. \
Please refer to the online documentation or run 'mikado util class_codes for details."},
            "validate": validate.ContainsOnly(["p", "P", "i", "I", "ri", "rI", "x", "X", "m", "_", "e", "o"])
        })


@dataclass
class FilesConfiguration:
    output_dir: str = field(default=".", metadata={
        "metadata": {"description": "Output directory for mikado pick"},
        "validate": validate.Length(min=1)
    })
    input: str = field(default="mikado_prepared.gtf", metadata={
        "metadata": {"description": "Input GTF/GFF3/BED12 file. Default: mikado_prepared.gtf"},
        "required": True
    })
    loci_out: str = field(default="mikado.loci.gff3", metadata={
        "metadata": {"description": "Main output GFF3 file from Mikado pick. Default: mikado.loci.gff3"},
    })
    subloci_out: Optional[str] = field(default=None, metadata={
        "metadata": {"description": "Optional GFF file with the intermediate subloci. Default: no output"},
    })
    monoloci_out: Optional[str] = field(default=None, metadata={
        "metadata": {"description": "optional GFF file with the intermediate monoloci. Default: no output"},
    })
    log: str = field(default="pick.log", metadata={
        "metadata": {"description": "Log file for mikado pick."}
    })


@dataclass
class PickConfiguration:
    scoring_file: str = field(default="plant.yaml", metadata={
        "metadata": {"description": "Scoring file to be used by Mikado."},
        "required": True
    })
    alternative_splicing: AlternativeSplicingConfiguration = field(
        default_factory=AlternativeSplicingConfiguration,
        metadata={"metadata": {
            "description": "Parameters related to how Mikado will select and report alternative splicing events."},
        })
    output_format: OutputFormatConfiguration = field(default_factory=OutputFormatConfiguration, metadata={
                "metadata": {"description": "Parameters related to the output format."},
    })
    orf_loading: OrfLoadingConfiguration = field(default_factory=OrfLoadingConfiguration, metadata={
                "metadata": {"description": "Parameters related to ORF loading."},
    })
    chimera_split: ChimeraSplitConfiguration = field(default_factory=ChimeraSplitConfiguration, metadata={
                "metadata": {
                    "description": "Parameters related to the splitting of transcripts \
in the presence of two or more ORFs."},
    })
    run_options: RunOptionsConfiguration = field(default_factory=RunOptionsConfiguration, metadata={
                "metadata": {"description": "Generic run options for Mikado pick."},
    })
    clustering: ClusteringConfiguration = field(default_factory=ClusteringConfiguration, metadata={
                "metadata": {"description": "Parameters related to the clustering of transcripts into loci."},
    })
    fragments: FragmentsConfiguration = field(default_factory=FragmentsConfiguration, metadata={
                "metadata": {"description": "Parameters related to the handling of fragments."}
    })
    files: Optional[FilesConfiguration] = field(default_factory=FilesConfiguration, metadata={
                "metadata": {"description": "Input and output files for Mikado pick."},
    })



# coding: utf-8

"""
This module defines the last object to be created during the picking,
i.e. the locus.
"""
import copy
from typing import Union, Dict, List, Set, Iterable
import collections
import itertools
import operator
from collections import defaultdict
import pysam

from ..scales.resultstorer import ResultStorer
from ..transcripts.pad import pad_transcript
from ..transcripts.transcript import Transcript
from .abstractlocus import Abstractlocus
from ..parsers.GFF import GffLine
from ..scales.assignment.assigner import Assigner
import networkx as nx
import random

from ..utilities import IntervalTree, Interval


class Locus(Abstractlocus):
    """Class that defines the final loci.
    It is a child of monosublocus, but it also has the possibility of adding
    additional transcripts if they are valid splicing isoforms.
    """

    def __init__(self, transcript=None, logger=None, configuration=None, **kwargs):
        """
        Constructor class. Like all loci, also Locus is defined starting from a transcript.

        :param transcript: the transcript which is used to initialize the Locus
        :type transcript: [Transcript|None]

        :param logger: the logger instance.
        :type logger: None | logging.Logger
        """

        self.counter = 0  # simple tag to avoid collisions
        Abstractlocus.__init__(self, logger=logger, configuration=configuration, **kwargs)
        if transcript is not None:
            transcript.attributes["primary"] = True
            if transcript.is_coding:
                transcript.feature = "mRNA"
            else:
                transcript.feature = "ncRNA"
            self.monoexonic = transcript.monoexonic
            Abstractlocus.add_transcript_to_locus(self, transcript)
            self.locus_verified_introns = transcript.verified_introns
            self.tid = transcript.id
            self.logger.debug("Created Locus object with {0}".format(transcript.id))
            self.primary_transcript_id = transcript.id

        # this must be defined straight away
        # A set of the transcript we will ignore during printing
        # because they are duplications of the original instance. Done solely to
        # get the metrics right.
        self._orf_doubles = collections.defaultdict(set)
        self.excluded = None
        self.parent = None

        self.attributes = dict()
        self.attributes["is_fragment"] = False
        self.metric_lines_store = []
        self.__id = None
        self.fai = None
        self.__finalized = False
        self._reference_sources = set(source for source, is_reference in
                                      zip(self.configuration.prepare.files.labels,
                                          self.configuration.prepare.files.reference) if is_reference is True)

        self.valid_ccodes = self.configuration.pick.alternative_splicing.valid_ccodes[:]
        self.redundant_ccodes = self.configuration.pick.alternative_splicing.redundant_ccodes[:]

        if self.perform_padding is True and self.reference_update is True:
            self._add_to_alternative_splicing_codes("=")
            self._add_to_alternative_splicing_codes("_")
            self._add_to_alternative_splicing_codes("n")

    def __str__(self, print_cds=True) -> str:

        self.feature = self.__name__
        assert self.feature != "Monosublocus"
        # Hacky fix to make sure that the primary transcript has the attribute
        # Set to True in any case.
        self.primary_transcript.attributes["primary"] = True
        # BF, really just a hack.
        for transcript in self.transcripts:
            if transcript == self.primary_transcript_id:
                continue
            self.transcripts[transcript].attributes["primary"] = False

        lines = []

        self_line = GffLine('')
        for attr in ["chrom", 'feature', 'source', 'start', 'end', 'strand']:
            setattr(self_line, attr, getattr(self, attr))
        self_line.phase, self_line.score = None, self.score
        self_line.id = self.id
        self_line.name = self.name
        self_line.attributes["superlocus"] = self.parent
        self_line.attributes.update(self.attributes)
        if "is_fragment" in self.attributes and self.attributes["is_fragment"] is False:
            del self_line.attributes["is_fragment"]
        self_line.attributes["multiexonic"] = (not self.monoexonic)
        lines.append(str(self_line))

        for tid in self.transcripts:
            transcript_instance = self.transcripts[tid]
            transcript_instance.source = self.source
            transcript_instance.parent = self_line.id
            self.logger.debug(self.attributes)
            if self.transcripts[tid].is_coding:
                self.transcripts[tid].feature = "mRNA"
            else:
                self.transcripts[tid].feature = "ncRNA"
            for attribute in self.attributes:
                if attribute not in transcript_instance.attributes:
                    if attribute == "is_fragment" and self.attributes[attribute] is False:
                        continue
                    transcript_instance.attributes[attribute] = self.attributes[attribute]

            lines.append(transcript_instance.format(
                "gff", with_cds=print_cds,
                all_orfs=self.configuration.pick.output_format.report_all_orfs
            ).rstrip())

        return "\n".join(lines)

    def __setstate__(self, state):
        super(Locus, self).__setstate__(state)
        self._not_passing = set(self._not_passing)

    def finalize_alternative_splicing(self, _scores=None, check_requirements=True):

        """"This method ensures that all the transcripts retained in the locus
        are within the score threshold. This is due to the fact that the score
        changes depending on the transcript considered together; so that a transcript
        that might have scored relatively well on its own will score pretty badly when
        brought inside the locus.

        Please note that, **when doing a reference update**, reference transcripts that have passed the previous checks
        will *always* be retained and will not count for the purposes of determining the maximum number of isoforms.

        :param _scores:only for testing. Load directly a score mock-up.

        :param check_requirements: boolean switch. If set to False, requirements will not be evaluated.
        """

        if self._finalized is True:
            self.logger.debug("Locus already finalised, returning")
            return
        else:
            self.logger.debug("Starting to finalise AS for %s", self.id)

        if _scores is not None:
            self.logger.warning("Loadin external scores: %s", _scores)
            self._load_scores(_scores)
        else:
            self.metrics_calculated = False
            self.scores_calculated = False
            self.logger.debug("Re-calculating metrics and scores for %s", self.id)
            self.filter_and_calculate_scores(check_requirements=check_requirements)
            self.logger.debug("Re-calculated metrics and scores for %s", self.id)

        max_isoforms = self.configuration.pick.alternative_splicing.max_isoforms
        original = {tid: self.transcripts[tid].copy() for tid in self.transcripts}

        # *Never* lose the primary transcript
        reference_sources = {source for source, is_reference in zip(
                self.configuration.prepare.files.labels,
                self.configuration.prepare.files.reference
            ) if is_reference is True}
        reference_transcripts = {
            tid: self.transcripts[tid].is_reference
            or self.transcripts[tid].original_source in reference_sources
            for tid in self.transcripts
        }

        order = sorted([(tid, self.transcripts[tid].score, self.transcripts[tid]) for tid in self.transcripts
                        if tid != self.primary_transcript_id and
                        (reference_transcripts[tid] is False or
                         self.configuration.pick.run_options.reference_update is False)],
                       key=operator.itemgetter(1), reverse=True)
        threshold = self.configuration.pick.alternative_splicing.min_score_perc * self.primary_transcript.score

        score_passing = [_ for _ in order if _[1] >= threshold]
        removed = {_[0] for _ in order if _[1] < threshold}

        remainder = score_passing[max_isoforms:]
        score_passing = score_passing[:max_isoforms]

        self.remove_transcripts_from_locus(removed)
        self.remove_transcripts_from_locus([tup[0] for tup in remainder])

        self.logger.debug("Kept %s transcripts out of %s (threshold: %s)",
                          len(score_passing), len(order), threshold)

        iteration = 0
        while True:
            iteration += 1
            self.logger.debug("Starting iteration %s", iteration)
            if _scores is None:
                self.metrics_calculated = False
                self.scores_calculated = False
                self.filter_and_calculate_scores(check_requirements=check_requirements)
            with_retained = self._remove_retained_introns()
            if len(with_retained) > 0:
                self.logger.debug("Transcripts with retained introns: %s", ", ".join(list(with_retained)))
            else:
                self.logger.debug("No transcripts with retained introns found.")
            if self.perform_padding is True and len(self.transcripts) > 1:
                self.logger.debug("Starting padding procedure for %s", self.id)
                failed = self.launch_padding()
                if failed:
                    # Restart the padding procedure
                    continue
                removed.update(self._remove_redundant_after_padding())
                self.primary_transcript.attributes.pop("ccode", None)
                self.logger.debug("Updated removal set: %s", ", ".join(removed))
            missing = len(self.transcripts) - max_isoforms
            if missing <= 0 or not remainder:
                break
            __to_add = remainder[:missing]
            for tid, score, obj in __to_add:
                self.add_transcript_to_locus(obj, check_in_locus=False)
            remainder = remainder[missing:]
        self._mark_padded_transcripts(original)
        # Now that we have added the padding ... time to remove redundant alternative splicing events.
        self.logger.debug("%s has %d transcripts (%s)", self.id, len(self.transcripts),
                          ", ".join(list(self.transcripts.keys())))
        self.metrics_calculated = False
        self.scores_calculated = False
        self.filter_and_calculate_scores(check_requirements=check_requirements)
        assert self.primary_transcript_id in self._metrics, (self.primary_transcript_id,
                                                             self._metrics.keys(),
                                                             self.transcripts.keys())
        assert self.primary_transcript_id in self.scores, (self.primary_transcript_id,
                                                           self.scores.keys(),
                                                           self.transcripts.keys())
        self._finalized = True
        return

    def _mark_padded_transcripts(self, original: dict):
        for tid in self.transcripts:
            # For each transcript check if they have been padded
            assert tid in original
            backup = original[tid]
            transcript = self.transcripts[tid]
            if (transcript.start, transcript.end) != (backup.start, backup.end):
                self.transcripts[tid].attributes["padded"] = True
                message = (f"{transcript.id} is now padded and has now start "
                           "{transcript.start}, end {transcript.end}")
                if backup.is_coding:
                    if (
                        backup.combined_cds_end != transcript.combined_cds_end
                        or backup.combined_cds_start
                        != transcript.combined_cds_start
                    ):
                        transcript.attributes["cds_padded"] = True
                        message += (f"; CDS moved to {transcript.combined_cds_start}, "
                                    "end {transcript.combined_cds_end}")
                    else:
                        transcript.attributes["cds_padded"] = False
                message += "."
                self.logger.info(message)

    def launch_padding(self):

        """Method to launch the padding procedure."""

        self.logger.debug("Launched padding for %s", self.id)
        failed = False
        backup = {tid: self.transcripts[tid].deepcopy() for tid in self.transcripts}
        backup_graph = nx.DiGraph(incoming_graph_data=self.internal_graph)
        l = []
        self.segmenttree.traverse(l.append)
        backup_segmenttree = IntervalTree.from_intervals(l)
        self.logger.warning(f"Checking the graph for {backup.keys()}")
        # The "templates" are the transcripts that we used to expand the others.
        templates = self.pad_transcripts(backup=None)
        # First off, let us update the transcripts.
        tid_keys = list(self.transcripts.keys())
        for tid in tid_keys:
            self.logger.debug("Swapping %s", tid)
            self._swap_transcript(backup[tid], self.transcripts[tid])

        self.logger.debug("Done padding for %s", self.id)
        self.metrics_calculated = False
        self.scores_calculated = False
        self.filter_and_calculate_scores()
        self.logger.debug("Recalculated metrics after padding in %s", self.id)

        self._not_passing = set()
        self._check_requirements()  # Populate self._not_passing with things that fail the requirements
        # If we would have to remove the primary transcript we have done something wrong
        if self.primary_transcript_id in self._not_passing or len(set.intersection(templates, self._not_passing)) > 0:
            self.logger.warning(
                "Either the primary or some template transcript has not passed the muster. Removing, restarting.")
            if self._not_passing == {self.primary_transcript_id}:
                self.logger.warning(
                    "The primary transcript %s is invalidated by the other transcripts in the locus. "
                    "Leaving only the main transcript in %s.", self.primary_transcript_id, self.id)
                self._not_passing = set(self.transcripts.keys()) - {self.primary_transcript_id}
            # Remove transcripts *except the primary* that do not pass the muster
            self.remove_transcripts_from_locus(self._not_passing - {self.primary_transcript_id})
            # Restore the *non-failed* transcripts *and* the primary transcript to their original state
            for tid in set(backup.keys()) - (self._not_passing - {self.primary_transcript_id}):
                self._swap_transcript(self.transcripts[tid], backup[tid])
            self.metrics_calculated = False
            self.scores_calculated = False
            self.filter_and_calculate_scores()
            # Signal to the master function that we have to redo the cycle
            failed = True
            return failed

        # Order the transcripts by score. The primary is *not* in this list
        order = sorted([(tid, self.transcripts[tid].score) for tid in self.transcripts
                        if tid != self.primary_transcript_id],
                       key=operator.itemgetter(1), reverse=True)

        self.logger.debug(
            "Transcripts potentially kept in the locus: %s",
            ",".join(_[0] for _ in order),
        )

        # Now that we are sure that we have not ruined the primary transcript, let us see whether
        # we should discard any other transcript.
        [self._add_to_alternative_splicing_codes(ccode) for ccode in ("=", "_")]
        removed = set()
        added = set()

        for tid, score in order:
            is_valid, ccode, _ = self.is_alternative_splicing(self.transcripts[tid], others=added)
            if is_valid:
                self.logger.debug("Keeping %s in the locus, ccode: %s", tid, ccode)
                self.transcripts[tid].attributes["ccode"] = ccode
                added.add(tid)
            else:
                self.logger.debug("Removing %s from the locus after padding, ccode: %s", tid, ccode)
                self.remove_transcript_from_locus(tid)
                removed.add(tid)

        removed.update(self._remove_retained_introns())

        # Now let us check whether we have removed any template transcript.
        # If we have, remove the offending ones and restart
        if len(set.intersection(set.union(set(templates), {self.primary_transcript_id}), removed)) > 0:
            self.logger.debug("Removed: %s; Templates: %s; Primary: %s", ",".join(removed), ",".join(templates),
                              self.primary_transcript_id)
            assert all(transcript.finalized is True for transcript in self.transcripts.values())
            assert all(transcript.finalized is True for transcript in backup.values())
            self.transcripts = backup
            assert backup_graph != self.internal_graph
            self._internal_graph = backup_graph
            assert self._internal_graph == backup_graph
            self.__segmenttree = backup_segmenttree
            self.logger.debug(f"Removing: {set.intersection(templates, removed)}")
            self.remove_transcripts_from_locus(set.intersection(templates, removed))
            self.metrics_calculated = False
            self.scores_calculated = False
            self.filter_and_calculate_scores()
            failed = True
            self.logger.debug("Padding failed for %s (removed: %s), restarting", self.id, removed)
            return failed

        return failed

    def _remove_retained_introns(self) -> Set[str]:
        """Method to remove from the locus any transcript that has retained introns and/or their CDS disrupted by
        a retained intron event."""

        self.logger.debug("Now checking the retained introns for %s", self.id)
        removed = set()
        while True:
            cds_disrupted = set()
            retained_introns = set()
            to_remove = set()
            for tid, transcript in self.transcripts.items():
                if tid == self.primary_transcript_id:
                    continue
                # This function will *only* modify the transcript instances, specifically *only* the
                # "cds_disrupted_by_ri" and "retained_introns_num" properties
                self.find_retained_introns(transcript)
                self.transcripts[tid].attributes["retained_intron"] = (transcript.retained_intron_num > 0)
                if transcript.retained_intron_num > 0:
                    retained_introns.add(tid)
                assert transcript.cds_disrupted_by_ri is False or transcript.retained_intron_num > 0
                if transcript.cds_disrupted_by_ri is True:
                    cds_disrupted.add(tid)
            if max(len(retained_introns), len(cds_disrupted)) == 0:
                break
            if self.configuration.pick.alternative_splicing.keep_cds_disrupted_by_ri is False and cds_disrupted:
                self.logger.debug("Removing {} because their CDS is disrupted by retained introns".format(
                    ", ".join(list(cds_disrupted))))
                to_remove.update(cds_disrupted)
                retained_introns -= cds_disrupted
            if self.configuration.pick.alternative_splicing.keep_retained_introns is False and retained_introns:
                self.logger.debug("Removing {} because they contain retained introns".format(
                    ", ".join(list(retained_introns))))
                to_remove.update(retained_introns)
            if not to_remove:
                break

            self.logger.debug(f"Removing from {self.id}: {' '.join(to_remove)}")
            removed.update(to_remove)
            self.remove_transcripts_from_locus(to_remove)
            self.__segmenttree = self._calculate_segment_tree(self.exons, self.introns)
            self.metrics_calculated = False
            self.scores_calculated = False
            self.filter_and_calculate_scores()
        return removed

    def _remove_redundant_after_padding(self):

        """Private method to remove duplicate copies of transcripts after the padding procedure."""

        # First thing: calculate the class codes
        to_remove_redundant = set()
        if len(self.transcripts) == 1:
            self.logger.debug(f"{self.id} only has one transcript, no redundancy removal needed.")
            return to_remove_redundant

        self.logger.debug("Starting to remove redundant transcripts from %s", self.id)

        class_codes = dict()
        ichains = collections.defaultdict(list)

        for tid in self.transcripts:
            if self.transcripts[tid].introns:
                key = tuple(sorted(self.transcripts[tid].introns))
            else:
                key = None
            self.logger.debug("Intron key for %s: %s", tid, key)
            ichains[key].append(tid)

        for ichain in ichains:
            for t1, t2 in itertools.combinations(ichains[ichain], 2):
                class_codes[(t1, t2)] = Assigner.compare(self[t1], self[t2])[0]
                self.logger.debug("Comparing ichain %s, %s vs %s: nF1 %s", ichain, t1, t2, class_codes[(t1, t2)].n_f1)

        for couple, comparison in class_codes.items():
            if comparison.n_f1[0] == 100:
                if self.primary_transcript_id in couple:
                    removal = [_ for _ in couple if _ != self.primary_transcript_id][0]
                elif len([_ for _ in couple if self[_].is_reference]) >= 1:
                    removal = [_ for _ in couple if self[_].is_reference is False][0]
                elif self[couple[0]].score != self[couple[1]].score:
                    removal = sorted([(_, self[_].score) for _ in couple],
                                     key=operator.itemgetter(1), reverse=True)[1][0]
                else:
                    removal = random.choice(sorted(couple))
                if removal:
                    to_remove_redundant.add(removal)
                    self.logger.debug("Removing %s from locus %s because after padding it is redundant with %s",
                                      removal, self.id, (set(couple) - {removal}).pop())

        if to_remove_redundant:
            self.logger.debug("Removing from %s: %s", self.id, ", ".join(to_remove_redundant))
            self.remove_transcripts_from_locus(to_remove_redundant)

        [self._add_to_redundant_splicing_codes(_) for _ in ("=", "_", "n", "c")]
        [self._remove_from_alternative_splicing_codes(_) for _ in ("=", "_", "n", "c")]
        order = sorted([(tid, self[tid].score) for tid in self if tid != self.primary_transcript_id],
                       key=operator.itemgetter(1))
        others = set(self.transcripts.keys())

        to_remove_alt_splice = set()
        for tid, score in order:
            if self[tid].is_reference:
                continue
            is_valid, main_ccode, main_result = self.is_alternative_splicing(self[tid], others=others)
            if is_valid is False:
                self.logger.debug("Removing %s from %s as it is a redundant splicing isoform after padding.",
                                  tid, self.id)
                to_remove_alt_splice.add(tid)
                others.remove(tid)

        if to_remove_alt_splice:
            self.logger.debug("Removing from %s: %s", self.id, ", ".join(to_remove_alt_splice))
            self.remove_transcripts_from_locus(to_remove_alt_splice)

        return to_remove_redundant.union(to_remove_alt_splice)

    def remove_transcript_from_locus(self, tid: str):

        """Overloading of the AbstractLocus class, in order to ensure that the primary transcript will *not*
        be removed."""

        if tid == self.primary_transcript_id:
            raise KeyError("%s is the primary transcript of %s!" % (tid, self.id))

        super().remove_transcript_from_locus(tid)
        self._finalized = False

    def add_transcripts_to_locus(self, transcripts: Iterable[Transcript],
                                 check_in_locus=True, **kwargs):
        """Simple iteration of the add_transcript_to_locus method. This is *different* from the base class;
        for loci, we want to add every transcript one by one and check many additional things."""

        [self.add_transcript_to_locus(transcript, check_in_locus=check_in_locus)
         for transcript in transcripts]

    def add_transcript_to_locus(self, transcript: Transcript, check_in_locus=True,
                                **kwargs):
        """Implementation of the add_transcript_to_locus method.
        Before a transcript is added, the class checks that it is a valid splicing isoform
        and that we have not exceeded already the maximum number of isoforms for the Locus.

        The checks performed are, in order:

        #. whether the locus already has the maximum number of acceptable
           isoforms ("max_isoforms")
        #. (optional) whether all the introns *specific to the transcript when
           compared with the primary transcript* are confirmed by external
           validation tools (eg Portcullis)
        #. Whether the score of the proposed AS event has a score over the
           minimum percentage of the primary transcript score (eg if the minimum
           percentage is 0.6 and the primary is scored 20, a model with a score
           of 11 would be rejected and one with a score of 12 would be accepted)
        #. Whether the strand of the candidate is the same as the one of the locus
        #. Whether the AS event is classified (ie has a class code) which is acceptable as valid AS
        #. Whether the transcript shares enough cDNA with the primary transcript ("min_cdna_overlap")
        #. Whether the proposed model has too much UTR
        #. (optional) Whether the proposed model has a retained intron compared
           to the primary, ie part of its 3' non-coding regions overlaps
           one intron of the primary model
        #. Whether the proposed model shares enough CDS with the primary model (min_cds_overlap)

        :param transcript: the candidate transcript
        :type transcript: Transcript

        :param check_in_locus: boolean flag to bypass *all* checks. Use with caution!
        :type check_in_locus: bool

        :param kwargs: optional keyword arguments are ignored.
        """

        _ = kwargs
        if check_in_locus is False:
            Abstractlocus.add_transcript_to_locus(self, transcript, check_in_locus=False)
            self.locus_verified_introns.update(transcript.verified_introns)
            return

        to_be_added = True

        if to_be_added and transcript.strand != self.strand:
            self.logger.debug("%s not added because it has a different strand from %s (%s vs. %s)",
                              transcript.id, self.id, transcript.strand, self.strand)
            to_be_added = False

        reference_pass = False
        if self.reference_update is True and \
                (transcript.is_reference is True or transcript.original_source in self._reference_sources):
            reference_pass = True

        if (
            self.configuration.pick.alternative_splicing.only_confirmed_introns
            is True
            and not reference_pass
        ):
            to_check = (transcript.introns - transcript.verified_introns) - self.primary_transcript.introns
            to_be_added = len(to_check) == 0
            if not to_be_added:
                self.logger.debug(
                    "%s not added because it has %d non-confirmed intron%s",
                    transcript.id,
                    len(to_check),
                    "s" * min(1, len(to_check) - 1))

        # Add a check similar to what we do for the minimum requirements and the fragments
        if to_be_added and self.configuration.scoring.as_requirements:
            to_be_added = self._check_as_requirements(transcript, is_reference=reference_pass)

        if to_be_added is True:
            is_alternative, ccode, _ = self.is_alternative_splicing(transcript)
            if is_alternative is False:
                self.logger.debug("%s not added because it is not a \
                valid splicing isoform. Ccode: %s",
                                  transcript.id, ccode)
                to_be_added = False
            else:
                transcript.attributes["ccode"] = ccode
                self.logger.debug("%s is a valid splicing isoform; Ccode: %s", transcript.id, ccode)

        if to_be_added is False:
            return

        self.logger.debug("Keeping %s as a valid alternative isoform for %s",
                          transcript.id, self.id)
        transcript.attributes["primary"] = False

        transcript.feature = "mRNA" if transcript.is_coding else "ncRNA"
        Abstractlocus.add_transcript_to_locus(self, transcript)

        self.locus_verified_introns.update(transcript.verified_introns)

    def _check_as_requirements(self, transcript: Transcript, is_reference=False) -> bool:
        """Private method to evaluate a transcript for inclusion in the locus.
        This method uses the "as_requirements" section of the configuration file to perform the
        evaluation.
        Please note that if "check_references" is False and the transcript is marked as reference, this function
        will always evaluate to True (ie the transcript is valid).
        """

        to_be_added = True
        if is_reference is True and self.configuration.pick.run_options.check_references is False:
            return True

        section = self.configuration.scoring.as_requirements

        evaluated = dict()
        for key in section.parameters:
            name = section.parameters[key].name
            if 'attributes' in name:
                name_parts = name.split('.')
                default = section.parameters[key].default
                value = transcript.attributes.get(name_parts[1], default)
            else:
                value = operator.attrgetter(name)(transcript)
                if "external" in key:
                    value = value[0]

            evaluated[key] = self.evaluate(value, section.parameters[key])
            # pylint: disable=eval-used
        if eval(section.compiled) is False:
            self.logger.debug("%s fails the minimum requirements for AS events", transcript.id)
            to_be_added = False
        return to_be_added

    def is_intersecting(self, *args):
        """Not implemented: See the alternative splicing definition functions.
        """
        raise NotImplementedError("""Loci do not use this method, but rather assess whether a transcript is a 
        splicing isoform or not.""")

    def is_putative_fragment(self):

        """This method will use the expression in the "not_fragmentary" section
        of the configuration to determine whether it is itself a putative fragment."""

        if not self.configuration.pick.run_options.check_references and \
                any(self.transcripts[tid].is_reference is True for tid in self.transcripts):
            return False

        current_id = self.id[:]

        evaluated = dict()
        for key, params in self.configuration.scoring.not_fragmentary.parameters.items():
            name = params.name
            if name.startswith("attributes"):
                default = params.default
                value = self.primary_transcript.attributes.get(name.replace("attributes.", ""), default)
            else:
                value = operator.attrgetter(name)(self.primary_transcript)
            if "external" in key:
                value = value[0]
            try:
                evaluated[key] = self.evaluate(value, params)
            except Exception as err:
                self.logger.error(
                    """Exception while calculating putative fragments. Key: {}, \
                    Transcript value: {} (type {}) \
                    configuration value: {} (type {}).""".format(
                        key, value, type(value), params,
                        type(params)
                    ))
                self.logger.exception(err)
                raise err
        if eval(self.configuration.scoring.not_fragmentary.compiled) is True:
            self.logger.debug("%s cannot be a fragment according to the definitions, keeping it",
                              self.id)
            fragment = False
        else:
            self.logger.debug(
                "%s could be a fragment according to the definitions, tagging it for analysis",
                self.id)
            fragment = True

        self.id = current_id
        assert self.id == current_id
        return fragment

    def other_is_fragment(self, other) -> (bool, Union[None, ResultStorer]):
        """
        If the 'other' locus is marked as a potential fragment (see 'is_putative_fragment'), then this function
        will check whether the other locus is within the distance and with the correct comparison class code to be
        marked as a fragment.

        :param other: another Locus to compare against
        :type other: Locus
        """

        if not isinstance(self, type(other)):
            raise TypeError("I can compare only loci.")

        if other.primary_transcript_id == self.primary_transcript_id:
            self.logger.debug("Self-comparisons are not allowed!")
            return False, None

        if other.is_putative_fragment() is False:
            self.logger.debug(f"{other.id} cannot be a fragment according to the scoring file definition.")
            return False, None

        if any(other.transcripts[tid].is_reference is True for tid in other.transcripts.keys()):
            self.logger.debug("Locus %s has a reference transcript, hence it will not be discarded", other.id)
            return False, None

        self.logger.debug("Comparing %s with %s",
                          self.primary_transcript_id,
                          other.primary_transcript_id)

        result, _ = Assigner.compare(other.primary_transcript,
                                     self.primary_transcript,
                                     strict_strandedness=True)

        # We should get rid of the "max_distance" value here.
        max_distance = max(0,
                           min(self.configuration.pick.fragments.max_distance,
                               self.configuration.pick.clustering.flank))

        self.logger.debug("Comparison between {0} (strand {3}) and {1}: class code \"{2}\"".format(
            self.primary_transcript.id,
            other.primary_transcript.id,
            result.ccode[0],
            other.strand))
        if (result.ccode[0] in self.configuration.pick.fragments.valid_class_codes and
                    result.distance[0] <= max_distance):
            self.logger.debug("{0} is a fragment (ccode {1})".format(
                other.primary_transcript.id, result.ccode[0]))
            return True, result

        return False, None

    def calculate_metrics(self, tid: str):
        """
        This function will calculate the metrics for a transcript which are relative in nature
        i.e. that depend on the other transcripts in the sublocus. Examples include the fraction
        of introns or exons in the sublocus, or the number/fraction of retained introns.

        :param tid: the name of the transcript to be analysed
        :type tid: str
        """

        self.logger.debug("Calculating metrics for %s", tid)
        self.transcripts[tid].finalize()
        if (self.transcripts[tid].number_internal_orfs <= 1 or
                    self.configuration.pick.output_format.report_all_orfs is False):
            super().calculate_metrics(tid)
        else:
            super().calculate_metrics(tid)
            transcript = self.transcripts[tid]
            orfs = list(transcript.get_internal_orf_beds())
            selected = orfs[transcript.selected_internal_orf_index]
            template = transcript.copy()
            template.strip_cds()
            new_transcript = template.copy()
            self.logger.debug("Changing the name of %s to %s", transcript.id,
                              ", ".join([transcript.id + ".orf" + str(_)
                                         for _ in range(1, len(transcript.internal_orfs) + 1)]))

            new_transcript.id = "{0}.orf1".format(new_transcript.id)
            from ..parsers.bed12 import BED12
            assert isinstance(selected, BED12), (type(selected), selected)
            new_transcript.load_orfs([selected])

            self.transcripts[new_transcript.id] = new_transcript
            super().calculate_metrics(new_transcript.id)
            if tid not in self._orf_doubles:
                self._orf_doubles[tid] = set([])

            self._orf_doubles[tid].add(new_transcript.id)

            for num, orf in enumerate([_ for _ in orfs if _ != selected]):
                new_transcript = template.copy()
                assert isinstance(new_transcript, Transcript)
                new_transcript.load_orfs([orf])
                new_transcript.id = "{0}.orf{1}".format(new_transcript.id, num + 2)
                self.transcripts[new_transcript.id] = new_transcript
                super().calculate_metrics(new_transcript.id)
                self._orf_doubles[tid].add(new_transcript.id)

        self.logger.debug("Calculated metrics for {0}".format(tid))

    def filter_and_calculate_scores(self, check_requirements=True):
        """
        Function to calculate a score for each transcript, given the metrics derived
        with the calculate_metrics method and the scoring scheme provided in the JSON configuration.
        If any requirements have been specified, all transcripts which do not pass them
        will be assigned a score of 0 and subsequently ignored.
        Scores are rounded to the nearest integer.
        """

        self.get_metrics()
        metrics_store = self._metrics.copy()
        if len(self._orf_doubles) == 0:
            doubled = set()
        else:
            doubled = set.union(*self._orf_doubles.values())
        for key, item in metrics_store.items():
            if item["tid"] in doubled:
                del self._metrics[key]
                self.transcripts.pop(item["tid"], None)

        super().filter_and_calculate_scores(check_requirements=check_requirements)

        self._metrics = metrics_store
        self.logger.debug("Calculated scores for %s, now checking for double IDs", self.id)

        for index, item in enumerate(reversed(self.metric_lines_store)):
            if item["tid"] in self._orf_doubles:
                del self.metric_lines_store[index]
            else:
                continue

        for doubled in self._orf_doubles:
            for partial in self._orf_doubles[doubled]:
                if partial in self.transcripts:
                    del self.transcripts[partial]
                if partial in self.scores:
                    del self.scores[partial]
        assert self.metrics_calculated is True
        assert self.scores_calculated is True

    def print_metrics(self):
        """Overloading of the base method as in loci we might want to print data for the double ORFs."""

        if self.configuration.pick.output_format.report_all_orfs is False:
            yield from super().print_metrics()
        else:
            self.get_metrics()
            ignore = set.union(*self._orf_doubles.values())
            for tid, transcript in sorted(self.transcripts.items(), key=operator.itemgetter(1)):
                if tid in ignore:
                    continue
                yield self._create_metrics_row(tid, self._metrics[tid], transcript)
                founds = []
                if tid in self._orf_doubles:
                    for mtid in self._orf_doubles[tid]:
                        founds.append(mtid)
                        yield self._create_metrics_row(mtid, self._metrics[mtid], transcript)
                elif transcript.alias in self._orf_doubles:
                    for mtid in self._orf_doubles[transcript.alias]:
                        founds.append(mtid)
                        yield self._create_metrics_row(mtid, self._metrics[mtid], transcript)

    def print_scores(self):
        """This method yields dictionary rows that are given to a csv.DictWriter class."""
        self.filter_and_calculate_scores()
        score_keys = sorted(list(self.configuration.scoring.scoring.keys()) + ["source_score"])
        keys = ["tid", "alias", "parent", "score"] + score_keys

        for tid in self.scores:
            row = dict().fromkeys(keys)
            row["tid"] = tid
            row["parent"] = self.id
            row["alias"] = self.transcripts[tid].alias
            if tid in self._not_passing:
                row["score"] = 0
            else:
                row["score"] = round(self.scores[tid]["score"], 2)
            for key in score_keys:
                assert self.scores[tid][key] != "NA" and self.scores[tid][key] is not None
                row[key] = round(self.scores[tid][key], 2)

            score_sum = sum(row[key] for key in score_keys)
            if tid not in self._not_passing and self.scores[tid]["score"] > 0:
                assert round(score_sum, 2) == round(self.scores[tid]["score"], 2), (
                    score_sum,
                    self.transcripts[tid].score,
                    tid)
            else:
                assert self.scores[tid]["score"] == 0

            yield row

    def is_alternative_splicing(self, other, others=None):

        """This function defines whether another transcript could be a
        putative alternative splice variant of the primary Locus
        transcript.
        To do so, it compares the candidate against all transcripts in the Locus, and calculates
        the class code using scales.Assigner.compare.

        :param other: another transcript to compare against
        :type other: Transcript

        :param others: a set of AS transcripts used to check whether the new transcript should be included.
        This allows to avoid creating a whole new locus for testing the padding.
        :type others: (None|set)
        """

        is_valid = True

        valid_ccodes = self.valid_ccodes
        redundant_ccodes = self.redundant_ccodes
        cds_only = self.configuration.pick.alternative_splicing.cds_only

        if other.is_coding and not self.primary_transcript.is_coding:
            reason = "{} is coding, and cannot be added to a non-coding locus.".format(other.id)
            enough_overlap, overlap_reason = False, reason
            main_ccode = "NA"
            main_result = None
        elif (other.is_coding is False) and (self.primary_transcript.is_coding is True):
            reason = "{} is non-coding, and cannot be added to a coding locus.".format(other.id)
            enough_overlap, overlap_reason = False, reason
            main_ccode = "NA"
            main_result = None
        else:
            if cds_only is True:
                self.logger.debug("Checking whether the CDS of %s and %s are overlapping enough",
                                  other.id, self.primary_transcript_id)
                main_result, _ = Assigner.compare(other._selected_orf_transcript,
                                                  self.primary_transcript._selected_orf_transcript)

                enough_overlap, overlap_reason = self._evaluate_transcript_overlap(
                    other._selected_orf_transcript,
                    self.primary_transcript._selected_orf_transcript,
                    min_cdna_overlap=self.configuration.pick.alternative_splicing.min_cdna_overlap,
                    min_cds_overlap=self.configuration.pick.alternative_splicing.min_cds_overlap,
                    comparison=main_result,
                    check_references=self.configuration.pick.run_options.check_references,
                    fixed_perspective=True)
            else:
                main_result, _ = Assigner.compare(other,
                                                  self.primary_transcript)
                enough_overlap, overlap_reason = self._evaluate_transcript_overlap(
                    other,
                    self.primary_transcript,
                    min_cdna_overlap=self.configuration.pick.alternative_splicing.min_cdna_overlap,
                    min_cds_overlap=self.configuration.pick.alternative_splicing.min_cds_overlap,
                    comparison=main_result,
                    check_references=self.configuration.pick.run_options.check_references,
                    fixed_perspective=True)

            self.logger.debug(overlap_reason)

            main_ccode = main_result.ccode[0]
            if main_ccode not in valid_ccodes:
                self.logger.debug("%s is not a valid splicing isoform. Ccode: %s",
                                  other.id,
                                  main_result.ccode[0])
                is_valid = False

        if not enough_overlap:
            self.logger.debug("%s is not a valid splicing isoform. Reason: %s",
                              other.id, overlap_reason)
            is_valid = False

        if (is_valid and self.configuration.pick.run_options.check_references is False and other.is_reference is True):
            pass
        elif is_valid:
            if others is None:
                others = self.transcripts.keys()
            else:
                diff = set.difference(set(others), set(self.transcripts.keys()))
                assert diff == set(), diff

            for tid in iter(tid for tid in others if tid not in (self.primary_transcript_id, other.id)):
                candidate = self.transcripts[tid]
                if cds_only is True:
                    result, _ = Assigner.compare(other._selected_orf_transcript, candidate._selected_orf_transcript)
                else:
                    result, _ = Assigner.compare(other, candidate)
                if result.ccode[0] in redundant_ccodes:
                    self.logger.debug("%s is a redundant isoform of %s (ccode %s)",
                                      other.id, candidate.id, result.ccode[0])
                    is_valid = False
                    break

        if is_valid:
            self.logger.debug("%s is a valid splicing isoform of %s (class code %s, overlap: %s)",
                              other.id, self.id, main_ccode, overlap_reason)

        return is_valid, main_ccode, main_result

    def pad_transcripts(self, backup=None) -> set:

        """
        This method will perform the padding of the transcripts.
        In a nutshell:
        - First, check which transcripts are compatible at the 5' and 3' end. Assign a "template" for each expansion.
        Note that the same transcript might be the template at the 5' but marked as extendable at the 3', or viceversa.
        - Call "expand_transcript" on each couple of template-target.

        Arguments:
            :param backup: optional dictionary of hard-copies of the transcripts to pad. Will be generated on the fly
            if None is provided.
        """

        if isinstance(self.configuration.reference.genome, pysam.FastaFile):
            self.fai = self.configuration.reference.genome
        else:
            self.fai = pysam.FastaFile(self.configuration.reference.genome)
        five_graph = self.define_graph(objects=self.transcripts, inters=self._share_extreme, three_prime=False)
        three_graph = self.define_graph(objects=self.transcripts, inters=self._share_extreme, three_prime=True)

        self.logger.debug("5' graph: %s", five_graph.edges)
        self.logger.debug("3' graph: %s", three_graph.edges)
        # TODO: Tie breaks!

        __to_modify = self._find_communities_boundaries(five_graph, three_graph)

        self.logger.debug("To modify: %s",
                          dict((tid, [_ if not isinstance(_, Transcript) else _.id for _ in __to_modify[tid]]) for tid in __to_modify)
                          )
        templates = set()
        if backup is None:
            backup = dict((tid, self.transcripts[tid].deepcopy()) for tid in self.transcripts)

        # Now we can do the proper modification
        for tid in sorted(__to_modify.keys()):
            if __to_modify[tid][0]:
                templates.add(__to_modify[tid][0].id)
            if __to_modify[tid][1]:
                templates.add(__to_modify[tid][1].id)

            self.logger.debug("Expanding %s to have start %s (from %s) and end %s (from %s)",
                              tid, __to_modify[tid][0] if not __to_modify[tid][0] else __to_modify[tid][0].start,
                              self[tid].start,
                              __to_modify[tid][1] if not __to_modify[tid][1] else __to_modify[tid][1].end,
                              self[tid].end)
            new_transcript = pad_transcript(backup[tid],
                                            self.transcripts[tid],
                                            __to_modify[tid][0],
                                            __to_modify[tid][1],
                                            self.fai,
                                            self.logger)
            if (new_transcript.start == self.transcripts[tid].end) and \
                    (new_transcript.end == self.transcripts[tid].end):
                self.logger.debug("No expansion took place for %s!", tid)
            else:
                self.logger.debug("Expansion took place for %s!", tid)
            self.transcripts[tid] = new_transcript

        self.exons = set()
        for tid in self:
            self.exons.update(self[tid].exons)
        return templates

    def _find_communities_boundaries(self, five_graph, three_graph) -> Dict[str, List[Union[bool, Transcript]]]:

        """This private method will navigate the 5' and 3' graph to assign a template for expansion to each target.
        It returns a dictionary where each transcript in the locus is assigned a bi-tuple - one item is the
        template at the 5', the other the template at the 3'.
        If no valid template is found, the corresponding item will be the boolean False.
        """

        five_found = set()

        __to_modify = dict()

        while len(five_graph) > 0:
            # Find the sinks
            sinks = {node for node in five_graph.nodes() if node not in
                     {edge[0] for edge in five_graph.edges()}}
            __putative = defaultdict(list)
            for sink in sinks:
                for ancestor in nx.ancestors(five_graph, sink):
                    __putative[ancestor].append((sink, self[sink].score))

            for ancestor in __putative:
                best = sorted(__putative[ancestor], key=operator.itemgetter(1), reverse=True)[0][0]
                self.logger.debug("Putative 5' for %s: %s. Best: %s", ancestor, __putative[ancestor], best)
                __to_modify[ancestor] = [self[best], False]
                five_found.add(ancestor)

            five_graph.remove_nodes_from(set.union(sinks, __putative.keys()))

        three_found = set()
        while len(three_graph) > 0:
            sinks = {node for node in three_graph.nodes() if node not in
                     {edge[0] for edge in three_graph.edges()}}
            __putative = defaultdict(list)
            for sink in sinks:
                for ancestor in nx.ancestors(three_graph, sink):
                    __putative[ancestor].append((sink, self[sink].score))

            for ancestor in __putative:
                best = sorted(__putative[ancestor], key=operator.itemgetter(1), reverse=True)[0][0]
                if ancestor in __to_modify:
                    __to_modify[ancestor][1] = self[best]
                else:
                    __to_modify[ancestor] = [False, self[best]]
                three_found.add(ancestor)
                self.logger.debug("Putative 3' for %s: %s. Best: %s", ancestor, __putative[ancestor], best)
            three_graph.remove_nodes_from(set.union(sinks, __putative.keys()))

        self.logger.debug("Communities for modifications: %s",
                          dict((tid, [_ if not isinstance(_, Transcript) else _.id for _ in __to_modify[tid]])
                               for tid in __to_modify))

        return __to_modify

    def define_graph(self, objects: dict, inters=None, three_prime=False) -> nx.DiGraph:

        """Method to determine the internal graph representation to be used to determine padding templates.

        :param objects: the dictionary containing the transcripts
        :param inters: the "is_intersecting" function to use (must evaluate to boolean). If None is provided, the
        method _share_extreme will be used.
        :param three_prime: whether to construct the graph for the 5' (False) or 3' (True) ending.
        """

        graph = nx.DiGraph()
        graph.add_nodes_from(objects.keys())
        inters = self._share_extreme if inters is None else inters

        if len(objects) >= 2:
            if (three_prime is True and self.strand != "-") or (three_prime is False and self.strand == "-"):
                reverse = True
            else:
                reverse = False
            order = sorted([(objects[tid].start, objects[tid].end, tid) for tid in objects], reverse=reverse)

            for pos in range(len(order) - 1):
                obj = order[pos]
                for other_obj in order[pos + 1:]:
                    if obj == other_obj:
                        continue
                    elif self.overlap(obj[:2], obj[:2], positive=False, flank=0) == 0:
                        break
                    else:
                        self.logger.debug("Comparing %s to %s (%s')", obj[2], other_obj[2],
                                          "5" if not three_prime else "3")
                        edge = inters(objects[obj[2]], objects[other_obj[2]], three_prime=three_prime)
                        if edge:
                            assert edge[0].id in self
                            assert edge[1].id in self
                            graph.add_edge(edge[0].id, edge[1].id)
        else:

            self.logger.debug("No comparison to be made (objects: %s)", objects)

        return graph

    def _share_extreme(self, first: Transcript, second: Transcript, three_prime=False):

        """
        This function will determine whether two transcripts "overlap" at the 3' or 5' end.
        The things to be considered are:
        - whether their extremes are within the maximum distance
        - whether unifying them would bridge too many splice sites within the locus.
        :param first:
        :param second:
        :return:
        """

        if self.strand == "-":  # Remember we have to invert on the negative strand!
            three_prime = not three_prime

        if three_prime:
            return self._share_three_prime(first, second)
        else:
            return self._share_five_prime(first, second)

    def _share_five_prime(self, first: Transcript, second: Transcript):

        """
        Method to determine whether two transcripts are compatible for expansion at the 5'.
        To be compatible:
        - their starts must be different
        - the number of splicing junctions to add must be at most the parameter "ts_max_splices"
        - the total number of bases to be added must be at most "ts_max_distance"

        :param first:
        :param second:
        :return:
        """


        reason = None
        ts_splices = 0
        ts_distance = 0

        if second.start == first.start:
            self.logger.debug("%s and %s start at the same coordinate. No expanding.", first.id, second.id)
            return False

        decision = False
        first, second = sorted([first, second], key=operator.attrgetter("start"))
        if self.overlap((first.start, first.end), (second.start, second.end)) <= 0:
            decision = False
            reason = "{first.id} and {second.id} are not intersecting each other."
            self.logger.debug(reason)
            return decision            

        # Now let us check whether the second falls within an intron
        matched = first.segmenttree.find(second.exons[0][0], second.exons[0][1])
        self.logger.debug("{second.id} last exon {second.exons[0]} intersects in {first.id}: {matched}".format(
            **locals()))
        if len(matched) > 0 and (matched[0].value == "intron" or second.exons[0][0] < matched[0].start):
            decision = False
            reason = "{second.id} first exon ends within an intron of {first.id}".format(**locals())
        elif len(matched) > 0:
            upstream = [_ for _ in first.find_upstream(second.exons[0][0], second.exons[0][1])
                        if _.value == "exon" and _ not in matched]
            if matched[0][0] < second.start:
                if upstream:
                    ts_splices += 1
                ts_distance += second.start - matched[0][0] + 1
            for up in upstream:
                if up.start == first.start:
                    ts_splices += 1
                else:
                    ts_splices += 2
                ts_distance += up.end - up.start - 1

        if reason is None:
            decision = (ts_distance <= self.ts_distance) and (ts_splices <= self.ts_max_splices)
            if decision:
                decision = (second, first)
            reason = "{first.id} {doesit} overlap {second.id} (distance {ts_distance} max {self.ts_distance}, splices \
{ts_splices} max {self.ts_max_splices})".format(doesit="does" if decision else "does not", **locals())
        self.logger.debug(reason)
        return decision

    def _share_three_prime(self, first: Transcript, second: Transcript):

        """
        Method to determine whether two transcripts are compatible for expansion at the 3'.
        To be compatible:
        - their starts must be different
        - the number of splicing junctions to add must be at most the parameter "ts_max_splices"
        - the total number of bases to be added must be at most "ts_max_distance"

        :param first:
        :param second:
        :return:
        """

        if second.end == first.end:
            self.logger.debug("%s and %s end at the same coordinate. No expanding.", first.id, second.id)
            return False

        reason = None
        ts_splices = 0
        ts_distance = 0
        decision = False
        first, second = sorted([first, second], key=operator.attrgetter("end"), reverse=False)
        # Now let us check whether the second falls within an intron
        if self.overlap((first.start, first.end), (second.start, second.end)) <= 0:
            decision = False
            reason = "{first.id} and {second.id} are not intersecting each other."
            self.logger.debug(reason)
            return decision            
        
        matched = second.segmenttree.find(first.exons[-1][0], first.exons[-1][1])
          
        if len(matched) > 0 and (matched[-1].value == "intron" or first.exons[-1][1] > matched[-1].end):
            decision = False
            reason = "{first.id} last exon ends within an intron of {second.id}".format(**locals())
        elif len(matched) > 0:
            downstream = [_ for _ in second.find_downstream(first.exons[-1][0], first.exons[-1][1])
                          if _.value == "exon" and _ not in matched]

            if matched[-1][1] > first.end:
                if downstream:
                    ts_splices += 1
                ts_distance += matched[-1][1] - first.end + 1

            for down in downstream:
                if down.end == second.end:
                    ts_splices += 1
                else:
                    ts_splices += 2
                ts_distance += down.end - down.start - 1

        if reason is None:
            decision = (ts_distance <= self.ts_distance) and (ts_splices <= self.ts_max_splices)
            if decision:
                decision = (first, second)
            reason = "{second.id} {doesit} overlap {first.id} (distance {ts_distance} max \
{self.ts_distance}, splices {ts_splices} max {self.ts_max_splices})".format(
                doesit="does" if decision else "does not", **locals())
        self.logger.debug(reason)
        return decision

    @property
    def __name__(self):
        if len(self.transcripts) == 0:
            return "locus"
        elif any(transcript.selected_cds_length > 0 for
                 transcript in self.transcripts.values()):
            return "gene"
        else:
            return "ncRNA_gene"

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        Override of the abstractlocus method.
        :rtype str
        """
        if self.__id is None:
            myid = Abstractlocus.id.fget(self)  # @UndefinedVariable

            if self.counter > 0:
                myid = "{0}.{1}".format(myid, self.counter)
            # self.__set_id(myid)
            self.__id = myid

        return self.__id

    # pylint: disable=arguments-differ
    @id.setter
    def id(self, string):
        """
        Override of the original method from AbstractLocus. This override allows to
        create proper IDs for the final annotation to be output by Mikado.
        :param string:
        :return:
        """

        self.logger.debug("Setting new ID for %s to %s", self.id, string)
        self.__set_id(string)
        self.__id = string

    def __set_id(self, string):

        if string == self.__id:
            return

        self.logger.debug("Changing the ID of %s to %s", self.__id, string)
        primary_id = "{0}.1".format(string)
        old_primary = self.primary_transcript.id
        self.logger.debug("Changing the ID of the primary transcript of %s from %s to %s", self.__id,
                          self.primary_transcript.id, primary_id)
        self.primary_transcript.attributes["alias"] = self.primary_transcript.id
        self.primary_transcript.id = primary_id
        self.transcripts[primary_id] = self.primary_transcript
        self.primary_transcript_id = primary_id
        assert self.transcripts[primary_id].selected_cds_introns == self.transcripts[old_primary].selected_cds_introns
        del self.transcripts[old_primary]
        self._orf_doubles[primary_id] = {
            _.replace(old_primary, primary_id)
            for _ in self._orf_doubles.pop(old_primary, set())
        }

        order = sorted([k for k in self.transcripts.keys() if k != primary_id],
                       key=lambda xtid: self.transcripts[xtid])

        mapper = {old_primary: primary_id}

        for counter, tid in enumerate(order):
            counter += 2
            self.transcripts[tid].attributes["alias"] = tid
            new_id = "{0}.{1}".format(string, counter)
            self.transcripts[tid].id = new_id
            self.transcripts[new_id] = self.transcripts.pop(tid)
            olds = self._orf_doubles.pop(tid, set())
            self._orf_doubles[new_id] = {
                _.replace(tid, new_id) for _ in self._orf_doubles.pop(tid, set())
            }

            news = self._orf_doubles.pop(tid, set())
            self.logger.debug("Changed the old ORF IDs for %s from %s to %s", tid, olds, news)
            assert self._orf_doubles[new_id] is not None
            mapper[tid] = new_id

        if self.scores_calculated is True:
            new_scores = {}
            for tid in mapper:
                if tid not in self.scores:
                    error = (f"ID {tid} not found in {self.id} scores! Available IDs:\n"
                            f"{', '.join(self.scores.keys())}\n"
                            f"Mapper: {mapper.items()}\n"
                            f"Primary transcript ID: {self.primary_transcript_id}\n"
                            f"Old primary: {old_primary}")
                    self.logger.critical(error)
                    raise KeyError(error)
                values = self.scores[tid].copy()
                values["tid"] = mapper[tid]
                values["alias"] = tid
                new_scores[mapper[tid]] = values
            self.scores = new_scores

        if self.metrics_calculated is True:
            new_metrics = {}
            for tid in mapper:
                values = self._metrics[tid].copy()
                values["tid"] = mapper[tid]
                values["alias"] = tid
                new_metrics[mapper[tid]] = values
            self._metrics = new_metrics

    # pylint: enable=invalid-name,arguments-differ

    @property
    def is_fragment(self):
        """
        :rtype : bool
        Flag. It returns the value of self.attributes["is_fragment"]

        """
        return self.attributes["is_fragment"]

    @is_fragment.setter
    def is_fragment(self, val: bool):
        """
        Setter for is_fragment. Only boolean values are accepted.

        :param val: flag
        :type val: bool
        """
        if not isinstance(val, bool):
            raise ValueError(val)
        self.attributes["is_fragment"] = val

    @property
    def primary_transcript(self):
        """
        This property returns the primary transcript of the Locus
        (i.e. the one which has been used for creation and which has the highest score).
        :rtype : Transcript
        """

        return self.transcripts[self.primary_transcript_id]

    @property
    def purge(self):

        """Overloading of the base property. Loci should never purge."""

        return False

    @property
    def _finalized(self):
        return self.__finalized

    @_finalized.setter
    def _finalized(self, value):
        if not isinstance(value, bool):
            raise ValueError(value)
        self.__finalized = value

    @property
    def ts_distance(self):
        """Alias for self.configuration.pick.alternative_splicing.ts_distance"""
        return self.configuration.pick.alternative_splicing.ts_distance

    @property
    def ts_max_splices(self):
        """Alias for self.configuration.pick.alternative_splicing.ts_max_splices"""
        return self.configuration.pick.alternative_splicing.ts_max_splices

    @property
    def has_reference_transcript(self):
        """Checks whether any transcript in the locus is marked as reference."""
        return any(self.transcripts[transcript].is_reference for transcript in self)

    def _add_to_alternative_splicing_codes(self, code):
        """Method to retrieve the currently valid alternative splicing event codes"""
        codes = set(self.valid_ccodes)
        codes.add(code)
        self.valid_ccodes = list(codes)
        self._remove_from_redundant_splicing_codes(code)

    def _add_to_redundant_splicing_codes(self, code):
        """Method to retrieve the currently valid alternative splicing event codes"""
        codes = set(self.redundant_ccodes)
        codes.add(code)
        self.redundant_ccodes = list(codes)
        self._remove_from_alternative_splicing_codes(code)

    def _remove_from_alternative_splicing_codes(self, *ccodes):
        sub = self.valid_ccodes
        for ccode in ccodes:
            if ccode in sub:
                sub.remove(ccode)
        self.valid_ccodes = sub

    def _remove_from_redundant_splicing_codes(self, *ccodes):
        self.logger.debug("Removing from redundant ccodes: %s. Current: %s", ccodes,
                          self.redundant_ccodes)
        sub = self.redundant_ccodes
        sub = [_ for _ in sub if _ not in ccodes]
        self.logger.debug("New redundant ccodes: %s", sub)
        self.redundant_ccodes = sub



import os

from Mikado._transcripts.scoring_configuration import SizeFilter
from Mikado.scales.assignment import Assigner
import itertools
from Mikado import utilities
from Mikado.parsers.GFF import GffLine
from Mikado.loci import Transcript
from ..loci.locus import Locus, pad_transcript
from ..parsers.bed12 import BED12, Bed12Parser
from ..subprograms.util.trim import trim_coding, trim_noncoding
import unittest
import pysam
import pkg_resources
from ..utilities.log_utils import create_null_logger, create_default_logger
from ..configuration.configurator import load_and_validate_config


__author__ = 'Luca Venturini'


class TestTrimming(unittest.TestCase):

    def test_wrong_cds(self):

        transcript = Transcript()
        transcript.chrom = "15"
        transcript.source = "protein_coding"
        transcript.start = 47631264
        transcript.end = 48051999

        exons = [(47631264, 47631416),
                 (47704590, 47704669),
                 (47762671, 47762742),
                 (47893062, 47893093),
                 (47895572, 47895655),
                 (48051942, 48051999)]

        transcript.strand = "+"
        transcript.add_exons(exons)
        transcript.id = "ENST00000560636"
        transcript.parent = "ENSG00000137872"
        cds_line = "\t".join(["15",
                              "protein_coding",
                              "CDS",
                              "48051996",
                              "48051996",
                              ".",
                              "+",
                              "0",
                              "ID=ENST00000560636.cds1;Parent=ENST00000560636"])
        cds_line = GffLine(cds_line)
        transcript.add_exon(cds_line)
        logger = utilities.log_utils.create_null_logger()
        transcript.logger = logger
        with self.assertLogs("null", level="WARNING"):
            transcript.finalize()

        trimmed = trim_coding(transcript, logger, max_length=50)
        self.assertEqual(trimmed.start, 47631366)
        self.assertEqual(trimmed.end, 48051992)

    # @unittest.SkipTest
    def test_correct_cds(self):

        transcript = Transcript()
        transcript.chrom = "Chr1"
        transcript.source = "test"
        transcript.start = 10000
        transcript.end = 20000

        exons = [(10000, 11500),
                 (12000, 13000),
                 (15000, 18000),
                 (19000, 20000)]

        cds = [(11400, 11500),  # 101
               (12000, 13000),  # 1001 ==> 1102
               (15000, 17998)]  # 2998 == > 3090 (y)

        transcript.add_exons(exons)
        transcript.add_exons(cds, features="CDS")

        transcript.strand = "+"
        transcript.finalize()

        logger = utilities.log_utils.create_null_logger("correct_cds")

        copied = transcript.deepcopy()

        trimmed = trim_coding(copied, logger, max_length=50)
        self.assertEqual(trimmed.start, 11400)
        self.assertEqual(trimmed.end, 19050)

        copied = transcript.deepcopy()
        self.assertEqual(copied.start, 10000)
        trimmed = trim_coding(copied, logger, max_length=200)
        self.assertEqual(trimmed.start, 11300)
        self.assertEqual(trimmed.end, 19200)

    def test_noncoding(self):

        logger = utilities.log_utils.create_null_logger("correct_cds2")
        transcript = Transcript(logger=logger)
        transcript.chrom = "Chr1"
        transcript.source = "test"
        transcript.start = 10000
        transcript.end = 20000

        exons = [(10000, 11500),
                 (12000, 13000),
                 (15000, 18000),
                 (19000, 20000)]

        transcript.add_exons(exons)
        transcript.strand = "+"
        transcript.finalize()

        copied = transcript.deepcopy()

        trimmed = trim_noncoding(copied, max_length=50)
        self.assertEqual(trimmed.start, 11450)
        self.assertEqual(trimmed.end, 19050)

        copied = transcript.deepcopy()

        trimmed = trim_noncoding(copied, max_length=200)
        self.assertEqual(trimmed.start, 11300)
        self.assertEqual(trimmed.end, 19200)


class TestPadding(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.fai = pkg_resources.resource_filename("Mikado.tests", "chr5.fas.gz")

    def setUp(self):
        self.reference = "Chr5\t26574999\t26578625\tID=AT5G66600.3;coding=True;phase=0\t0\t-\t26575104\t26578315\t0\t11\t411,126,87,60,100,809,126,72,82,188,107\t0,495,711,885,1035,1261,2163,2378,2856,3239,3519"
        self.reference = Transcript(BED12(self.reference), source="TAIR10", is_reference=True)

    def test_basic_padding(self):

        logger = create_null_logger("test_basic_padding")
        logger.setLevel("INFO")
        template = self.reference.copy()
        template.id = "AT5G66600.3_exp"
        template.strip_cds()
        template.unfinalize()

        template.remove_exon((26575000, 26575410))  # First exon
        template.start = 26574650
        template.add_exon((26574970, 26575410))  # New exon, template at 5'
        template.add_exon((26574650, 26574820))  # New UTR exon

        template.remove_exon((26578519, 26578625))  # Last exon
        template.end = 26579700
        template.add_exon((26578519, 26578725))
        template.add_exon((26579325, 26579700))
        
        template.finalize()
        fai = pysam.FastaFile(pkg_resources.resource_filename("Mikado.tests", "chr5.fas.gz"))

        new5 = pad_transcript(self.reference, self.reference.deepcopy(), None, template, fai, logger)
        self.assertIn((26574970, 26575410), new5.exons)
        self.assertIn((26574650, 26574820), new5.exons)
        self.assertEqual(template.start, new5.start)
        self.assertEqual(self.reference.end, new5.end)

        new3 = pad_transcript(self.reference, self.reference.deepcopy(), template, None, fai, logger)
        self.assertIn((26578519, 26578725), new3.exons)
        self.assertIn((26579325, 26579700), new3.exons)
        self.assertEqual(self.reference.start, new3.start)
        self.assertEqual(template.end, new5.end)

        new53 = pad_transcript(self.reference, self.reference.deepcopy(), template, template, fai, logger)
        self.assertIn((26574970, 26575410), new53.exons)
        self.assertIn((26574650, 26574820), new53.exons)
        self.assertIn((26578519, 26578725), new53.exons)
        self.assertIn((26579325, 26579700), new53.exons)
        self.assertEqual(template.start, new53.start)
        self.assertEqual(template.end, new53.end)

    def test_locus_padding_equal_or_n(self):

        for num, exons_to_add in enumerate([
            ((26574970, 26575410), (26578519, 26578725)),
            ((26574970, 26575410), (26574650, 26574820),
             (26578519, 26578725), (26579325, 26579700))]):

            for num2, pad_transcripts in enumerate((False, True)):
                with self.subTest(exons_to_add=exons_to_add, pad_transcripts=pad_transcripts):
                    logger = create_null_logger("test_locus_padding_equal_or_n_" + str(num + num2 * 2))
                    logger.setLevel("DEBUG")
                    template = self.reference.copy()
                    del template.is_reference
                    template.id = "AT5G66600.3_exp"
                    template.unfinalize()
                    template.remove_exon((26575000, 26575410))  # First exon
                    template.remove_exon((26578519, 26578625))  # Last exon
                    template.start = min([_[0] for _ in exons_to_add])
                    template.end = max([_[1] for _ in exons_to_add])
                    template.add_exons(exons_to_add)  # New exon, template at 5'
                    template.finalize()
                    json_conf = load_and_validate_config(None)
                    json_conf.reference.genome = self.fai
                    json_conf.pick.alternative_splicing.only_confirmed_introns = False
                    json_conf.pick.run_options.only_reference_update = True
                    json_conf.pick.alternative_splicing.pad = pad_transcripts
                    locus = Locus(self.reference.copy(), logger=logger, configuration=json_conf)
                    self.assertTrue(locus[self.reference.id].is_reference)
                    self.assertEqual(locus.perform_padding, pad_transcripts)
                    locus.add_transcript_to_locus(template)
                    if pad_transcripts is True:
                        self.assertIn(template.id, locus)
                    locus.finalize_alternative_splicing()
                    self.assertNotIn(template.id, locus)
                    if pad_transcripts is False:
                        self.assertEqual(locus[self.reference.id].start, self.reference.start)
                        self.assertEqual(locus[self.reference.id].end, self.reference.end)
                    else:
                        self.assertTrue(locus.perform_padding)
                        self.assertEqual(locus[self.reference.id].start, template.start,
                                         (locus[self.reference.id].exons[0], template.exons[0]))
                        self.assertEqual(locus[self.reference.id].end, template.end,
                                         (locus[self.reference.id].end, template.end))
                        self.assertNotIn(template.id, locus)

    def test_removal_after_padding(self):

        """Here we test that, given three transcripts, the first one will be expanded to be identical to the second;
        the second will be removed as redundant; the third will be expanded to compatible with the padded first.
        """

        logger = create_default_logger("test_add_two_partials", "INFO")
        json_conf = load_and_validate_config(None)
        json_conf.reference.genome = self.fai
        json_conf.pick.alternative_splicing.min_cds_overlap = 0.2
        json_conf.pick.alternative_splicing.min_cdna_overlap = 0.2
        json_conf.pick.alternative_splicing.only_confirmed_introns = False
        json_conf.pick.alternative_splicing.keep_retained_introns = True
        json_conf.pick.alternative_splicing.pad = True
        json_conf.scoring.requirements.expression = ["cdna_length"]
        json_conf.scoring.requirements.parameters = {"cdna_length": SizeFilter(operator="gt", value=0)}
        json_conf.scoring.requirements._expression = json_conf.scoring.requirements._create_expression(
            json_conf.scoring.requirements.expression,
            json_conf.scoring.requirements.parameters)
        json_conf.scoring.as_requirements.expression = ["cdna_length"]
        json_conf.scoring.as_requirements.parameters = {"cdna_length": SizeFilter(operator="gt", value=0)}
        json_conf.scoring.as_requirements._expression = json_conf.scoring.requirements._create_expression(
            json_conf.scoring.requirements.expression,
            json_conf.scoring.requirements.parameters)

        t1 = Transcript(BED12(
            "Chr5\t26584779\t26587869\tID=AT5G66610.1;coding=True;phase=0\t0\t+\t26585222\t26587755\t0\t11\t\
100,54,545,121,78,105,213,63,119,59,443\t0,440,565,1202,1437,1640,1858,2154,2304,2507,2647"
        ))

        t2_1 = Transcript(BED12(
            "Chr5\t26584773\t26586510\tID=AT5G66610.2_1;coding=True;phase=0\t0\t+\t26585222\t26586510\t0\t\
6\t177,54,545,121,78,85\t0,446,571,1208,1443,1652"
        ))
        t2_2 = Transcript(BED12(
            "Chr5\t26584873\t26587782\tID=AT5G66610.2_2;coding=True;phase=0\t0\t+\t26585222\t\
26587755\t0\t10\t77,54,545,121,78,99,213,63,119,496\t0,346,471,1108,1343,1552,1764,2060,2210,2413"
        ))

        t1.finalize()
        t2_1.finalize()
        t2_2.finalize()
        t1.is_reference = True
        self.assertEqual(t1.start, 26584780)
        locus = Locus(t1, logger=logger, configuration=json_conf)
        locus.add_transcript_to_locus(t2_1, check_in_locus=False)
        locus.add_transcript_to_locus(t2_2, check_in_locus=False)
        self.assertTrue(locus.primary_transcript_id == t1.id)
        locus.logger.setLevel("INFO")
        locus.finalize_alternative_splicing(_scores={t1.id: 20, t2_1.id: 15, t2_2.id: 10})
        self.assertIn(t1.id, locus.transcripts)
        if t2_1.id in locus.transcripts:
            for tid1, tid2 in itertools.combinations(locus.transcripts.keys(), 2):
                res, _ = Assigner.compare(locus[tid1], locus[tid2])
                print(tid1, tid2, res.ccode)
            self.assertNotIn(t2_1.id, locus.transcripts.keys(),
                             [(key, val.start, val.end) for key, val in locus.transcripts.items()])

        self.assertIn(t2_2.id, locus.transcripts, "\n".join(tr.format("bed12") for tr in locus))
        self.assertTrue(locus[t2_2.id].attributes["padded"])
        # self.assertTrue(locus[t1.id].attributes["padded"])
        self.assertGreaterEqual(t1.start, locus[t1.id].start, locus[t1.id].format("bed12"))
        self.assertEqual(locus[t2_2.id].start, locus[t1.id].start,
                         ((locus[t2_2.id].start, t1.start, t2_1.start, t2_2.start),
                          (locus[t2_2.id].end, t1.end, t2_1.end, t2_2.end)),
                         )
        self.assertEqual(locus[t1.id].end, locus[t2_2.id].end)

    def test_add_two_partials(self):

        logger = create_null_logger("test_add_two_partials")
        logger.setLevel("INFO")
        json_conf = load_and_validate_config(None)
        json_conf.reference.genome = self.fai
        json_conf.pick.alternative_splicing.only_confirmed_introns = False
        json_conf.pick.run_options.only_reference_update = True

        ref = Transcript(is_reference=True)
        ref.chrom, ref.strand, ref.id = "Chr5", "-", "AT5G66670.2"
        ref.add_exons([(26611258, 26612889)])
        ref.add_exons([(26611474, 26612700)], features=["CDS"])
        ref.finalize()
        self.assertTrue(ref.is_coding)

        # Chr5	TAIR10	mRNA	26611258	26612889	.	-	.	ID=AT5G66670.2;Parent=AT5G66670;Name=AT5G66670.2;index=1
        # Chr5	TAIR10	protein	26611474	26612700	.	-	.	ID=AT5G66670.2-Protein;Parent=AT5G66670.2;Name=AT5G66670.2;derives_from=AT5G66670.2
        # Chr5	TAIR10	three_prime_UTR	26611258	26611473	.	-	.	Parent=AT5G66670.2
        # Chr5	TAIR10	CDS	26611474	26612700	.	-	0	Parent=AT5G66670.2
        # Chr5	TAIR10	five_prime_UTR	26612701	26612889	.	-	.	Parent=AT5G66670.2
        # Chr5	TAIR10	exon	26611258	26612889	.	-	.	Parent=AT5G66670.2

        template1 = Transcript(is_reference=False)
        template1.chrom, template1.strand, template1.id = ref.chrom, ref.strand, ref.id + "_frag1"
        template1.add_exons(((26611116, 26611157), (26611258, 26612670)))
        template1.add_exons(((26611474, 26612670),), features=["CDS"])
        template1.finalize()
        self.assertTrue(template1.is_coding)

        template2 = Transcript(is_reference=False)
        template2.chrom, template2.strand, template2.id = ref.chrom, ref.strand, ref.id + "_frag2"
        template2.add_exons(((26611574, 26612889), (26613007, 26613403)))
        template2.add_exons(((26611574, 26612700),), features=["CDS"])
        template2.finalize()
        self.assertTrue(template2.is_coding)

        logger.setLevel("INFO")
        json_conf.pick.alternative_splicing.pad = True
        locus = Locus(ref, configuration=json_conf, logger=logger)
        locus.add_transcript_to_locus(template1)
        locus.add_transcript_to_locus(template2)
        self.assertIn(template2.id, locus)
        # self.assertIn(template1.id, locus)
        # locus.logger.setLevel("DEBUG")
        # for tid in locus:
        #     locus[tid].logger.setLevel("DEBUG")
        locus.finalize_alternative_splicing(check_requirements=False)
        self.assertTrue(locus._finalized)
        self.assertNotIn(template1.id, locus, "\n" + str(locus))
        self.assertNotIn(template2.id, locus, "\n" + str(locus))
        self.assertEqual(locus[ref.id].end, template2.end,
                         ((locus[ref.id].end, ref.end, template2.end, template1.end),
                         (locus[ref.id].start, ref.start, template2.start, template1.start))
                         )

    @unittest.skip
    def test_failed_expansion(self):
        logger = create_default_logger("test_failed_expansion", level="WARNING")
        raw = [Transcript(line, logger=logger) for line in Bed12Parser(
            open(pkg_resources.resource_filename("Mikado.tests", os.path.join("test_pick_pad", "fail.bed12"))))]
        transcripts = dict((_.id, _) for _ in raw)
        [_.finalize() for _ in transcripts.values()]
        template = transcripts["template"]  # 4535908	4540293
        candidate = transcripts["candidate"]  # 4536444	4540027
        backup = candidate.copy()
        fai = pysam.FastaFile(pkg_resources.resource_filename("Mikado.tests", os.path.join("test_pick_pad",
                                                                                           "failing_seq.fa.gz")))
        logger.setLevel("DEBUG")
        candidate.logger.setLevel("DEBUG")
        pad_transcript(candidate, backup, start_transcript=template, end_transcript=template, fai=fai, logger=logger)


if __name__ == "__main__":
    unittest.main()



__author__ = 'Luca Venturini'


import unittest
from ..loci import Transcript
from ..utilities.log_utils import create_default_logger


class TestMetricsEndDistances(unittest.TestCase):

    logger = create_default_logger("End")
    logger.setLevel("ERROR")

    def setUp(self):

        self.tr = Transcript()
        self.tr.logger = self.logger
        self.tr.start = 101
        self.tr.end = 10000
        self.tr.add_exons([(101, 300),
                           (501, 800),
                           (1001, 1200),
                           (1301, 2000),
                           (3501, 5000),
                           (5501, 6000),
                           (6201, 7000),
                           (7301, 7700),
                           (8201, 9000),
                           (9101, 9300),
                           (9501, 9700),
                           (9801, 10000)])
        self.tr.id = "test1"
        self.tr.parent = "test1.gene"

    def test_end_positive(self):

        self.tr.strand = "+"

        cds = [(1161, 1200),  # 40 % 3 == 1
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9130)]

        self.tr.add_exons(cds, features="CDS")
        self.tr.finalize()
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)
        self.assertEqual(self.tr.selected_cds_end,
                         9130)
        self.assertEqual(self.tr.end_distance_from_junction,
                         (9300 - 9131 + 1) + (9700 - 9501 + 1)
                         )
        self.assertEqual(self.tr.end_distance_from_tes,
                         (9300 - 9131 + 1) + (9700 - 9501 + 1) + (10000 - 9801 + 1)
                         )

        self.tr.strip_cds()
        self.assertEqual(len(self.tr.internal_orfs), 0, self.tr.internal_orfs)
        self.tr.finalized = False
        cds = [(1161, 1200),  # 40 % 3 == 1
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9300),  # 200 % 3 == 2
               (9501, 9690)  # 190 % 3 == 1
               ]
        self.tr.add_exons(cds, features="CDS")

        self.tr.finalize()
        self.assertEqual(self.tr.combined_cds_end,
                         9690)
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)

        self.assertEqual(self.tr.end_distance_from_junction,
                         (9700 - 9691 + 1)
                         )
        self.assertEqual(self.tr.end_distance_from_tes,
                         (9700 - 9691 + 1) + (10000 - 9801 + 1)
                         )

        self.tr.strip_cds()
        self.assertEqual(self.tr.combined_cds_end,
                         self.tr.selected_cds_end,
                         self.tr.combined_cds)
        self.assertEqual(self.tr.combined_cds_end,
                         None,
                         self.tr.combined_cds_end)

        self.tr.finalized = False
        cds = [(1161, 1200),  # 40 % 3 == 1
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9300),  # 200 % 3 == 2
               (9501, 9700),  # 200 % 3 == 2
               (9801, 9820),  # 20 % 2 == 2
               ]
        self.tr.add_exons(cds, features="CDS")

        self.tr.finalize()
        self.assertEqual(self.tr.combined_cds_end,
                         9820)
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)
        self.assertEqual(self.tr.end_distance_from_tes,
                         180)
        self.assertEqual(self.tr.end_distance_from_junction,
                         0)

    def test_end_negative(self):

        self.tr.strand = "-"

        # self.tr.add_exons([(101, 300),
        #                    (501, 800),
        #                    (1001, 1200),
        #                    (1301, 2000),
        #                    (3501, 5000),
        #                    (5501, 6000),
        #                    (6201, 7000),
        #                    (7301, 7700),
        #                    (8201, 9000),
        #                    (9101, 9300),
        #                    (9501, 9700),
        #                    (9801, 10000)])

        cds = [(1161, 1200),  # 40 % 3 == 1
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9130)]

        self.assertEqual(sum(x[1] - x[0] + 1 for x in cds) % 3, 0)

        self.tr.add_exons(cds, features="CDS")
        self.tr.finalize()
        self.assertTrue(self.tr.is_coding)
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)
        self.assertEqual(self.tr.selected_cds_end,
                         1161)
        self.assertEqual(self.tr.end_distance_from_junction,
                         (1161-1001) + (800-501+1),
                         (self.tr.end_distance_from_junction,
                          (1161-1001) + (800-501+1))
                         )
        self.assertEqual(self.tr.end_distance_from_tes,
                         self.tr.end_distance_from_junction + (300 - 101 + 1),
                         (self.tr.end_distance_from_tes,
                          self.tr.end_distance_from_junction + (300 - 101 + 1))
                         )

        self.tr.strip_cds()
        self.assertEqual(len(self.tr.internal_orfs), 0, self.tr.internal_orfs)
        self.tr.finalized = False
        cds = [(721, 800),
               (1001, 1200),  # 200 % 3 == 2
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9130),  # 200 % 3 == 2
               ]
        self.tr.add_exons(cds, features="CDS")

        self.tr.finalize()
        self.assertEqual(self.tr.combined_cds_end,
                         721)
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)

        self.assertEqual(self.tr.end_distance_from_junction,
                         (721-501),
                         (self.tr.end_distance_from_junction, (721-501))
                         )
        self.assertEqual(self.tr.end_distance_from_tes,
                         self.tr.end_distance_from_junction + (300 - 101 + 1),
                         (self.tr.end_distance_from_tes,
                         self.tr.end_distance_from_junction + (300 - 101 + 1))
                         )

        self.tr.strip_cds()
        self.assertEqual(self.tr.combined_cds_end,
                         self.tr.selected_cds_end,
                         self.tr.combined_cds)
        self.assertEqual(self.tr.combined_cds_end,
                         None,
                         self.tr.combined_cds_end)

        self.tr.finalized = False
        cds = [(161, 300),    # 140 % 3 == 2
               (501, 800),    # 300 % 3 == 0
               (1001, 1200),  # 200 % 3 == 2
               (1301, 2000),  # 700 % 3 == 1
               (3501, 5000),  # 1500 % 3 == 0
               (5501, 6000),  # 500 % 3 == 2
               (6201, 7000),  # 800 % 3 == 2
               (7301, 7700),  # 400 % 3 == 1
               (8201, 9000),  # 800 % 3 == 2
               (9101, 9130),  # 30 % 3 == 0
               ]

        self.assertEqual(sum((_[1] - _[0] +1) % 3 for _ in cds ) % 3, 0)
        self.tr.logger = self.logger
        self.tr.add_exons(cds, features="CDS")
        self.tr.finalize()
        self.assertEqual(self.tr.combined_cds_end,
                         161)
        self.assertEqual(self.tr.selected_cds_end,
                         self.tr.combined_cds_end)
        self.assertEqual(self.tr.end_distance_from_tes,
                         60)
        self.assertEqual(self.tr.end_distance_from_junction,
                         0)


if __name__ == "__main__":
    unittest.main()


# coding: utf-8

"""This initializer contains the base declaration for all the DB classes of the module."""
import functools
import logging
import sqlite3
from dataclasses import field

from marshmallow import validate
from marshmallow_dataclass import dataclass, Optional
from sqlalchemy import event
from sqlalchemy.engine import create_engine, Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import database_exists, create_database

Inspector = Inspector
DBBASE = declarative_base()


@dataclass
class DBConfiguration:
    db: str = field(default="mikado.db",
                    metadata={"metadata": {"description": "the DB to connect to. Required. Default: mikado.db"},
                              "required": True,
                              "validate": validate.Length(min=1)}
                    )
    dbtype: str = field(default="sqlite",
                        metadata={
                            "metadata": {
                                "description": "Type of DB to use. Choices: sqlite, postgresql, mysql. Default: sqlite."
                            },
                            "required": True,
                            "validate": validate.OneOf(["sqlite", "mysql", "postgresql"])
    })
    dbhost: str = field(default="localhost", metadata={
                "metadata": {"description": "Host of the database. Unused if dbtype is sqlite. Default: localhost"}
    })
    dbuser: Optional[str] = field(default=None, metadata={
                "metadata": {"description": "DB user. Unused if dbtype is sqlite. Default: empty string"}
    })
    dbpasswd: Optional[str] = field(default=None, metadata={
                "metadata": {"description": "DB password for the user. Unused if dbtype is sqlite. Default: null"}
    })
    dbport: Optional[int] = field(default=None, metadata={
                "metadata": {
                    "description": "Integer. It indicates the default port for the DB. Unused if dbtype is sqlite."
                }
    })


def create_connector(configuration, logger=None):
    """Creator function for the database connection. It necessitates the following information from
    the json_conf dictionary:

    - dbtype (one of sqlite, mysql, postgresql)
    - db (name of the database file, for sqlite, otherwise name of the database)

    If the database is MySQL/PostGreSQL, the method also requires:

    - dbuser
    - dbhost
    - dbpasswd
    - dbport

    These are controlled and added automatically by the json_utils functions.

    :param configuration: configuration dictionary

    :param logger: a logger instance
    :type logger: logging.Logger

    :rtype : MySQLdb.connect | sqlite3.connect | psycopg2.connect

    """

    if logger is None:
        # Create a default null handler
        logger = logging.Logger("null")
        logger.addHandler(logging.NullHandler())

    db_settings = configuration.db_settings
    assert isinstance(db_settings, DBConfiguration)

    func = None
    if db_settings.dbtype == "sqlite":
        if not database_exists("sqlite:///{}".format(db_settings.db)):
            logger.debug("No database found, creating a mock one")
            create_database("sqlite:///{}".format(db_settings.db))
        logger.debug("Connecting to %s", db_settings.db)
        func = sqlite3.connect(database=db_settings.db, check_same_thread=False)
    elif db_settings.dbtype in ("mysql", "postgresql"):
        if db_settings.dbpasswd != '':
            passwd = ":{0}".format(db_settings.dbpasswd)
        else:
            passwd = ''
        url = "{dialect}://{user}{passwd}@{host}:{port}/{db}".format(
            dialect=db_settings.dbtype,
            host=db_settings.dbhost,
            user=db_settings.dbuser,
            passwd=passwd,
            db=db_settings.db,
            port=db_settings.dbport
        )
        if database_exists(url) is False:
            create_database(url)

        if db_settings.dbtype == "mysql":
            import MySQLdb
            logger.debug("Connecting to MySQL %s", db_settings.db)
            func = MySQLdb.connect(host=db_settings.dbhost,
                                   user=db_settings.dbuser,
                                   passwd=db_settings.dbpasswd,
                                   db=db_settings.db,
                                   port=db_settings.dbport)
        elif db_settings.dbtype == "postgresql":
            import psycopg2
            logger.debug("Connecting to PSQL %s", db_settings.db)
            func = psycopg2.connect(
                host=db_settings.dbhost,
                user=db_settings.dbuser,
                password=db_settings.dbpasswd,
                database=db_settings.db,
                port=db_settings.dbport
            )
    else:
        raise ValueError("DB type not supported! {0}".format(db_settings.dbtype))
    return func


def connect(configuration, logger=None, **kwargs):
    """
    Function to create an engine to connect to a DB with, using the
    configuration inside the provided json_conf.
    :param configuration:
    :param logger:
    :return: sqlalchemy.engine.base.Engine
    """

    @event.listens_for(Engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        try:
            cursor.execute("PRAGMA foreign_keys=ON")
            cursor.execute("PRAGMA synchronous=OFF")
            cursor.execute("PRAGMA temp_store=MEMORY")
            cursor.execute("PRAGMA journal_mode=MEMORY")
            cursor.execute("PRAGMA count_changes=OFF")
        except sqlite3.OperationalError:
            pass
        finally:
            cursor.close()

    if configuration is None:
        return create_engine("sqlite:///:memory:", **kwargs)

    db_connection = functools.partial(create_connector, configuration, logger=logger)
    engine = create_engine("{0}://".format(configuration.db_settings.dbtype),
                           creator=db_connection, **kwargs)
    DBBASE.metadata.create_all(engine, checkfirst=True)

    return engine



import multiprocessing as mp
from ...utilities.namespace import Namespace
from .transmission import transmit_transcript
from ..assignment.assigner import Assigner
from ..assignment.distributed import Assigners, FinalAssigner
from .transmission import get_best_result
from ..reference_preparation.gene_dict import GeneDict
from ..resultstorer import ResultStorer
from ...transcripts import Gene
from ...parsers.GFF import GFF3
from ...parsers.GTF import GTF
from ...parsers.bed12 import Bed12Parser
from ...parsers.bam_parser import BamParser
from .parse_bam_prediction import parse_prediction_bam
from .parse_gff3_prediction import parse_prediction_gff3
from .parse_gtf_prediction import parse_prediction_gtf
from .parse_bed12 import parse_prediction_bed12
from ...parsers import parser_factory
import gzip
import csv
import itertools


def parse_prediction(args, index, queue_logger):

    """
    This function performs the real comparison between the reference and the prediction.
     It needs the following inputs:
    :param args: the Namespace with the necessary parameters
    :param genes: Dictionary with the reference genes, of the form
    dict[chrom][(start,end)] = [gene object]
    :param positions: Dictionary with the positions of the reference genes, of the form
    dict[chrom][IntervalTree]
    :param queue_logger: Logger
    :return:
    """

    # start the class which will manage the statistics
    if hasattr(args, "self") and args.self is True:
        args.prediction = parser_factory(args.reference.name)
    __found_with_orf = set()
    queue = mp.JoinableQueue(100)
    returnqueue = mp.JoinableQueue(100)

    queue_logger.info("Starting to parse the prediction")
    if args.prediction.__annot_type__ == BamParser.__annot_type__:
        annotator = parse_prediction_bam
    elif args.prediction.__annot_type__ == Bed12Parser.__annot_type__:
        annotator = parse_prediction_bed12
    elif args.prediction.__annot_type__ == GFF3.__annot_type__:
        annotator = parse_prediction_gff3
    elif args.prediction.__annot_type__ == GTF.__annot_type__:
        annotator = parse_prediction_gtf
    else:
        raise ValueError("Unsupported input file format")

    if args.processes > 1:
        log_queue = args.log_queue
        dargs = dict()
        doself = False
        for key, item in args.__dict__.items():
            if key in ("log_queue", "queue_handler"):
                continue
            elif key == "self":
                doself = item
            else:
                dargs[key] = item
        nargs = Namespace(default=False, **dargs)
        nargs.self = doself
        procs = [Assigners(index, nargs, queue, returnqueue, log_queue, counter)
                 for counter in range(args.processes)]
        [proc.start() for proc in procs]
        final_proc = FinalAssigner(index, nargs, returnqueue, log_queue=log_queue, nprocs=len(procs))
        final_proc.start()
        done = 0
        lastdone = 1
        rows = []
        for transcript in annotator(args, queue_logger):
            rows, done, lastdone, __found_with_orf = transmit_transcript(
                transcript, done=done, lastdone=lastdone, rows=rows, queue=queue, queue_logger=queue_logger,
                __found_with_orf=__found_with_orf)

        rows, done, lastdone, __found_with_orf = transmit_transcript(
            None, done=done, lastdone=lastdone, rows=rows, queue=queue, queue_logger=queue_logger,
            __found_with_orf=__found_with_orf, send_all=True)

        queue_logger.info("Finished parsing, %s transcripts in total", done)
        queue.join()
        queue.put("EXIT")
        [proc.join() for proc in procs]
        final_proc.join()
    else:
        assigner_instance = Assigner(index, args, printout_tmap=True, )
        done = 0
        for transcript in annotator(args, queue_logger):
            if transcript is None:
                continue
            done += 1
            assigner_instance.get_best(transcript)
        queue_logger.info("Finished parsing, %s transcripts in total", done + 1)
        assigner_instance.finish()


def parse_self(args, gdict: GeneDict, queue_logger):

    """
    This function is called when we desire to compare a reference against itself.
    :param args:
    :param gdict: gene MIDX database
    :param queue_logger:
    :return:
    """

    queue_logger.info("Performing a self-comparison.")

    if args.gzip is False:
        tmap_out = open("{0}.tmap".format(args.out), 'wt')
    else:
        tmap_out = gzip.open("{0}.tmap.gz".format(args.out), 'wt')
    tmap_rower = csv.DictWriter(tmap_out, ResultStorer.__slots__, delimiter="\t")
    tmap_rower.writeheader()

    for gid, gene in gdict.items():
        assert isinstance(gene, Gene)
        if len(gene.transcripts) == 1:
            continue

        combinations = itertools.combinations(gene.transcripts.keys(), 2)
        for combination in combinations:
            result, _ = Assigner.compare(gene.transcripts[combination[0]],
                                         gene.transcripts[combination[1]],
                                         fuzzy_match=args.fuzzymatch)
            tmap_rower.writerow(result.as_dict())

    queue_logger.info("Finished.")



"""
This module provides the functions needed to check a transcript for consinstency,
e.g. reliability of the CDS/UTR, sanity of borders, etc.
"""
import itertools

from ...utilities import Interval, IntervalTree
from ...utilities import overlap
from collections import defaultdict
import operator
from ...exceptions import InvalidCDS, InvalidTranscript

__author__ = 'Luca Venturini'


def _basic_final_checks(transcript):

    """
    Function that verifies minimal criteria of a transcript before finalising.
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :return:
    """

    _exons = transcript.exons

    if not _exons:
        if transcript._possibly_without_exons is True:
            transcript.logger.debug("Inferring that %s is a single-exon transcript")
            _exons = [(transcript.start, transcript.end)]
        elif len(transcript.combined_cds) == 0:
            exc=InvalidTranscript(
                "No exon defined for the transcript {0}. Aborting".format(transcript.id))
            transcript.logger.exception(exc)
            raise exc
        else:
            # Let us try to derive exons from CDS ...
            _exons = sorted([(int(exon[0]), int(exon[1])) for exon in transcript.combined_cds])
            if len(transcript.combined_utr) == 0:
                # Enlarge the terminal exons to include the starts
                if transcript.start is not None:
                    _exons[0] = (transcript.start, _exons[0][1])
                if transcript.end is not None:
                    _exons[-1] = (_exons[-1][0], transcript.end)
            else:
                utr = sorted([(int(exon[0]), int(exon[1])) for exon in transcript.combined_utr])
                before = [_ for _ in utr if _[0] < _exons[0][0]]
                after = [_ for _ in utr if _[1] > _exons[-1][1]]
                if (before and before[-1][1] >= _exons[0][0]) or (after and after[0][0] <= _exons[-1][1]):
                    exc = InvalidTranscript("Transcript {} has a mangled CDS/UTR annotation. Please revise it.")
                    transcript.logger.exception(exc)
                    raise exc
                if before and before[-1][1] == _exons[0][0] - 1:
                    _exons[0] = (before[-1][0], _exons[0][1])
                    before.pop()
                if after and after[0][0] == _exons[-1][1] + 1:
                    _exons[-1] = (_exons[-1][0], after[0][1])
                    after = after[1:]
                _exons = before + _exons + after

    transcript.logger.debug("Converting to tuples")
    _exons = [(int(exon[0]), int(exon[1])) for exon in _exons]

    # Set the start and end automatically if none has been explicitly provided
    transcript._set_exons(sorted(_exons))
    # Now check for overlapping exons
    for exon_a, exon_b in zip(transcript.exons[:-1], transcript.exons[1:]):
        if exon_a[1] >= exon_b[0]:
            exc = InvalidTranscript(
                "Overlapping exons found for\n{0} {1}/{2}\n{3}".format(
                    transcript.id, exon_a, exon_b, transcript.exons))
            transcript.logger.debug(exc)
            raise exc

    _verify_boundaries(transcript)
    if len(transcript.exons) > 1 and transcript.strand is None and transcript._accept_undefined_multi is False:
        exc = InvalidTranscript(
            f"Multiexonic transcripts must have a defined strand! Error for {transcript.id}")
        transcript.logger.exception(exc)
        raise exc

    if transcript.combined_utr != [] and transcript.combined_cds == []:
        exc = InvalidTranscript(f"Transcript {transcript.id} has defined UTRs but no CDS feature!")
        transcript.logger.exception(exc)
        raise exc


def _check_cdna_vs_utr(transcript):

    """
    Verify that cDNA + UTR in the transcript add up.
    :return:
    """

    transcript.logger.debug("Checking the cDNA for %s", transcript.id)
    cdna_length = transcript.cdna_length
    if cdna_length > transcript.combined_utr_length + transcript.combined_cds_length:
        if transcript.combined_utr == transcript.combined_cds == []:
            # non-coding transcript
            transcript.logger.debug("%s is non coding, returning", transcript.id)
            return
        assert transcript.combined_cds != []

        transcript.logger.debug("Recalculating the UTR for %s. Reason: cDNA length %s, UTR %s, CDS %s (total %s)",
                                transcript.id, cdna_length, transcript.combined_utr_length,
                                transcript.combined_cds_length,
                                transcript.combined_utr_length + transcript.combined_cds_length)
        transcript.combined_utr = []  # Reset
        transcript.combined_cds = sorted(transcript.combined_cds,
                                         key=operator.itemgetter(0, 1))

        exons = IntervalTree.from_intervals([Interval(*exon) for exon in transcript.exons])

        mapper = defaultdict(list)
        for cds in transcript.combined_cds:
            fexon = exons.find(cds[0] - 1, cds[1], strict=False)
            if len(fexon) > 1:
                raise InvalidCDS(
                    "{} has a CDS ({}) which straddles {} different exons ({}).".format(
                        transcript.id, cds, len(fexon), fexon
                    )
                )
            elif len(fexon) == 0:
                raise InvalidCDS(
                    "{} has a CDS ({}) which is not mapped to any exon.".format(
                        transcript.id, cds, len(fexon), fexon
                    )
                )
            mapper[fexon[0]].append(cds)

        for exon in transcript.exons:
            if exon not in mapper:
                transcript.combined_utr.append(exon)
                continue
            elif len(mapper[exon]) == 1:
                cds = mapper[exon][0]
                if cds[0] == exon[0] and exon[1] == cds[1]:
                    continue
                else:
                    before = None
                    after = None
                    if cds[0] < exon[0] or cds[1] > exon[1]:
                        raise InvalidCDS(
                            "{} in {} debords its exon {}".format(cds, transcript.id, exon)
                        )
                    if cds[0] > exon[0]:
                        before = (exon[0], max(cds[0] - 1, exon[0]))
                        transcript.combined_utr.append(before)
                    if cds[1] < exon[1]:
                        after = (min(cds[1] + 1, exon[1]), exon[1])
                        transcript.combined_utr.append(after)
                    assert before or after, (exon, cds)
            else:
                transcript.logger.debug("Starting to find the UTRs for %s", exon)
                found = sorted(mapper[exon])
                utrs = []
                for pos, interval in enumerate(found):
                    if pos == len(found) - 1:
                        if exon[1] > interval[1]:
                            utrs.append((min(exon[1], interval[1] + 1), exon[1]))
                        continue
                    if pos == 0 and exon[0] < interval[0]:
                        utrs.append((exon[0], max(exon[0], interval[0] - 1)))
                    next_interval = found[pos + 1]
                    if not (interval[1] + 1 <= next_interval[0] - 1):
                        raise InvalidCDS(
                            "Error while inferring the UTR for a transcript with multiple ORFs: overlapping CDS found.")
                    utrs.append((interval[1] + 1, next_interval[0] - 1))
                assert utrs, found
                utr_sum = sum([_[1] - _[0] + 1 for _ in utrs])
                cds_sum = sum(_[1] - _[0] + 1 for _ in found)
                assert utr_sum + cds_sum == exon[1] - exon[0] + 1, (utr_sum, cds_sum,
                                                                    exon[1] - exon[0] + 1, utrs, found)
                transcript.combined_utr.extend(utrs)

        # If no CDS and no UTR are present, all good
        equality_one = (transcript.combined_cds_length == transcript.combined_utr_length == 0)
        # Otherwise, if cDNA length == UTR + CDS, all good
        equality_two = (cdna_length ==
                        transcript.combined_utr_length + transcript.combined_cds_length)
        if not (equality_one or equality_two):
            # Something fishy going on
            raise InvalidCDS(
                """"Failed to create the UTR:
ID: {}
Exons: {}
Combined CDS: {}
Combined UTR: {}
CDS == UTR == 0: {}
CDNA == CDS + UTR: {}
CDNA == {}
CDS == {}
UTR == {}""".format(transcript.id,
                    transcript.exons,
                    transcript.combined_cds,
                    transcript.combined_utr, equality_one, equality_two,
                    transcript.cdna_length, transcript.combined_cds_length, transcript.combined_utr_length))


def _calculate_introns(transcript):

    """Private method to create the stores of intron
    and splice sites positions.
    """

    introns = set()
    cds_introns = set()
    splices = set()

    if len(transcript.exons) > 1:
        for exon_a, exon_b in zip(transcript.exons[:-1], transcript.exons[1:]):
            # Append the splice junction
            introns.add((exon_a[1] + 1, exon_b[0] - 1))
            # Append the splice locations
            splices.update({exon_a[1] + 1, exon_b[0] - 1})
    transcript.introns = introns
    transcript.splices = splices

    if (transcript.number_internal_orfs == 0 or
            len(transcript.selected_cds) < 2 or
            len(transcript.combined_cds) < 2):
        pass
    else:
        # Start calculating the selected CDS introns
        for first, second in zip(transcript.selected_cds[:-1],
                                 transcript.selected_cds[1:]):
            assert first != second, (transcript.id, transcript.selected_cds)
            first, second = sorted([first, second])
            intron = tuple([first[1] + 1, second[0] - 1])
            if intron not in transcript.introns:
                continue
            cds_introns.add(intron)

        assert len(cds_introns) > 0
        transcript._selected_cds_introns = cds_introns

        if transcript.number_internal_orfs > 1:
            cds_introns = set()
            for position in range(len(transcript.combined_cds) - 1):
                former = transcript.combined_cds[position]
                latter = transcript.combined_cds[position + 1]
                junc = tuple([former[1] + 1, latter[0] - 1])
                if junc in transcript.introns:
                    cds_introns.add(junc)
            assert len(cds_introns) > 0
            transcript._combined_cds_introns = cds_introns
        else:
            transcript._combined_cds_introns = transcript._selected_cds_introns.copy()

    assert len(transcript.combined_cds_introns) >= len(transcript.selected_cds_introns)
    return transcript


def _check_completeness(transcript):

    """Private method that checks whether a transcript is complete
    or not based solely on the presence of CDS/UTR information."""

    if len(transcript.combined_utr) > 0:
        if transcript.combined_utr[0][0] < transcript.combined_cds[0][0]:
            if transcript.strand == "+":
                transcript.has_start_codon = True
            elif transcript.strand == "-":
                transcript.has_stop_codon = True
        if transcript.combined_utr[-1][1] > transcript.combined_cds[-1][1]:
            if transcript.strand == "+":
                transcript.has_stop_codon = True
            elif transcript.strand == "-":
                transcript.has_start_codon = True


def _verify_boundaries(transcript):

    """
    Method to verify that the start/end of the transcripts are exactly where they should.
    Called from finalise.
    :return:
    """

    if transcript.exons[0][0] != transcript.start or transcript.exons[-1][1] != transcript.end:
        transcript.logger.warning(
            """The transcript {id} has coordinates {tstart}:{tend},
        but its first and last exons define it up until {estart}:{eend}!
        Exons: {exons}. Modifying it""".format(
                id=transcript.id,
                tstart=transcript.start,
                tend=transcript.end,
                estart=transcript.exons[0][0],
                eend=transcript.exons[-1][1],
                exons=transcript.exons))
        transcript.start = transcript.exons[0][0]
        transcript.end = transcript.exons[-1][1]


def _calculate_phases(coding: list, previous: int) -> (int, dict):
    """

    :param coding:
    :param previous:
    :return:
    """

    total_cds_length = -previous

    __calculated_phases = dict()
    for cds_segment in coding:
        length = cds_segment[1][1] - cds_segment[1][0] + 1
        phase = (3 - (total_cds_length % 3)) % 3
        __calculated_phases[cds_segment[1]] = phase
        total_cds_length += length

    return total_cds_length, __calculated_phases


def _check_internal_orf(transcript, index):

    """
    Method that verifies that an internal ORF does not have any internal gap.

    :param transcript: the transcript to analyse
    :type transcript: Mikado.loci.Transcript
    :param index: index of the internal orf to check
    :type index: int

    :return: the updated transcript
    :rtype: Mikado.loci.Transcript
    """

    if transcript._trust_orf is True and index == 0:
        if (transcript.is_coding and transcript.phases) or not transcript.is_coding:
            new_orf = []
            for segment in transcript.internal_orfs[index]:
                if segment[0] == "CDS":
                    segment = tuple([segment[0], segment[1], transcript.phases[segment[1]]])
                new_orf.append(segment)
            transcript.internal_orfs[index] = new_orf
            return transcript
        else:
            pass

    orf, new_orf = transcript.internal_orfs[index], []

    exons = sorted(transcript.exons, reverse=(transcript.strand == "-"))

    coding = sorted([_ for _ in orf if _[0] == "CDS"], key=operator.itemgetter(1))
    transcript.logger.debug("ORF for %s: %s", transcript.id, coding)

    if not coding:
        err = "No ORF for {}, index {}!".format(transcript.id, index)
        transcript.logger.debug(err)
        raise InvalidCDS(err)

    before = sorted([_ for _ in orf
                     if _[0] == "UTR" and _[1][1] < coding[0][1][0]], key=operator.itemgetter(1))
    after = sorted([_ for _ in orf
                    if _[0] == "UTR" and _[1][0] > coding[-1][1][1]], key=operator.itemgetter(1))

    first = min(coding[0][1][0], float("inf") if not before else before[0][1][0])
    last = max(coding[-1][1][1], float("-inf") if not after else after[-1][1][1])

    if first != transcript.start or last != transcript.end:
        # Let's try to salvage this.
        err = """Invalid start and stop of the ORF for {}
First: {} Start: {}
Last: {} End {}
Coding: {}
Before: {}
After: {}
dict: {}""".format(transcript.id,
                   first, transcript.start,
                   last, transcript.end,
                   coding,
                   before,
                   after,
                   {},
                   # transcript.__dict__
                   )
        transcript.logger.debug(err)
        raise InvalidCDS(err)

    # Check that the number of exons with a coding section is correct and that they are in the correct order.
    coding_exons = [_ for _ in enumerate(exons) if
                    _[1][1] >= coding[0][1][1] and _[1][0] <= coding[-1][1][0]]
    if len(coding_exons) != len(coding) or coding_exons[-1][0] - coding_exons[0][0] + 1 != len(coding):
        raise InvalidCDS(""""Invalid number of coding exons for {} ({} vs {})
Coding: {}
Coding_exons (recalculated): {}""".format(
            transcript.id,
            len(coding), len(coding_exons),
            coding, coding_exons))

    # Now it's time to check the phases
    if transcript.strand == "-":
        coding = list(reversed(coding))
        five_utr = list(reversed(after))
        three_utr = list(reversed(before))
    else:
        five_utr = before
        three_utr = after

    del before, after

    phase_orf = []
    for _ in coding:
        if len(_) == 3:
            if _[2] not in (None, 0, 1, 2):
                raise ValueError("Invalid phase value for {}".format(transcript.id))
            phase_orf.append(_[2])
        elif len(_) == 2:
            continue
        else:
            raise ValueError("Invalid CDS fragment: {}".format(_))

    if len(phase_orf) != 0 and len(phase_orf) != len(coding):
        transcript.logger.warning("Invalid phases for %s. Resetting.", transcript.id)
        phase_orf = []

    if not phase_orf and transcript.phases:
        phases_keys = sorted(transcript.phases.keys(), reverse=(transcript.strand == "-"))
        phase_orf = [transcript.phases[_] for _ in phases_keys]
        # Calculating the complement of the phase so that
        # previous = (3 - phase_orf[0]) % 3
        previous = phase_orf[0]
        # transcript.logger.warning(previous)
    elif not phase_orf and transcript._first_phase is not None:
        previous = transcript._first_phase
        phase_orf = []
    elif phase_orf:
        previous = phase_orf[0]
    else:
        phase_orf = []
        for segment in sorted(orf, key=operator.itemgetter(1), reverse=(transcript.strand == "-")):
            if segment[0] != "CDS":
                continue
            else:
                if len(segment) == 3:
                    phase_orf.append(segment[2])
                else:
                    break
        if phase_orf and len(phase_orf) == len(coding):
            previous = phase_orf[0]
        else:
            previous = 0
            phase_orf = []

    total_cds_length, __calculated_phases = _calculate_phases(coding, previous)
    new_phases_keys = sorted(__calculated_phases.keys(), reverse=(transcript.strand == "-"))
    new_phase_orf = [__calculated_phases[_] for _ in new_phases_keys]

    assert len(__calculated_phases) == len(coding)

    if phase_orf and new_phase_orf != phase_orf:
        transcript.logger.debug("Wrong phases for %s, using recalculated ones (\n%s\nvs\n%s)",
                                transcript.id,
                                phase_orf, __calculated_phases)
    else:
        transcript.logger.debug("Correct phases for %s: %s",
                                transcript.id, __calculated_phases)
    if total_cds_length % 3 != 0 and three_utr and five_utr:
        # The transcript is truncated.
        raise InvalidCDS(""""Both UTR presents with a truncated ORF (length {}, modulo {}) in {};
5'UTR: {}
3' UTR: {}""".format(total_cds_length, total_cds_length % 3, transcript.id, five_utr, three_utr))
    elif total_cds_length % 3 != 0 and three_utr:
        for num in (0, 1, 2):
            total_cds_length, __calculated_phases = _calculate_phases(coding,
                                                                      num)
            if total_cds_length % 3 == 0:
                break

        if total_cds_length % 3 != 0:
            raise InvalidCDS("Persistently wrong ORF for %s at 5' end", transcript.id)

    if ((__calculated_phases[sorted(__calculated_phases.keys(), reverse=(transcript.strand == "-"))[0]] != 0)
            and five_utr):
        raise InvalidCDS("5'UTR present with a truncated ORF at 5' end for {}".format(
                         transcript.id))

    transcript.phases = __calculated_phases

    transcript.logger.debug("Total CDS length %d", total_cds_length)

    new_orf = five_utr[:]
    new_orf.extend([(_[0][0], _[0][1], _[1]) for _ in zip(
        coding,
        [__calculated_phases[_]  for _ in sorted(__calculated_phases.keys(), reverse=(transcript.strand == "-"))])])
    new_orf.extend(three_utr)

    new_orf.extend([("exon", _) for _ in transcript.exons])
    new_orf = sorted(new_orf, key=operator.itemgetter(1, 0))

    transcript.internal_orfs[index] = new_orf
    return transcript


def _check_phase_correctness(transcript, strip_faulty_cds=True):

    """
    This method verifies that the phases are assigned correctly in the case of a coding transcript.
    :param transcript: the input transcript.
    :type transcript: Mikado.loci.transcript.Transcript
    :return: Mikado.loci.transcript.Transcript
    """

    internal_orfs = transcript.internal_orfs

    segments = [("exon", tuple([e[0], e[1]])) for e in transcript.exons]
    # Define CDS
    if len(internal_orfs) > 0:
        for segment in itertools.chain(internal_orfs):
            if segment[0] == "exon":
                continue
            elif segment[0] == "UTR":
                segments.append(("UTR", (segment[1][0], segment[1][1])))
            elif segment[0] == "CDS":
                segments.append(("CDS", (segment[1][0], segment[1][1])))
    else:
        segments.extend([("CDS", tuple([c[0], c[1]])) for c in transcript.combined_cds])
    # Define UTR segments
    segments.extend([("UTR", tuple([u[0], u[1]])) for u in transcript.combined_utr])
    # Mix and sort
    segments = sorted(segments, key=operator.itemgetter(1, 0))
    # Add to the store as a single entity
    if not internal_orfs and any(_[0] == "CDS" for _ in segments):
        internal_orfs = [segments]
    else:
        transcript.selected_internal_orf_index = None

    transcript.internal_orfs = internal_orfs

    __orfs_to_remove = []
    orf_excs = []
    for orf_index in range(len(internal_orfs)):
        try:
            transcript = _check_internal_orf(transcript, orf_index)
        except (InvalidTranscript, InvalidCDS) as exc:
            transcript.logger.debug("ORF %s of %s is invalid. Exception: %s", orf_index, transcript.id, exc)
            __orfs_to_remove.append(orf_index)
            orf_excs.append(str(exc))

    __num_orfs = len(internal_orfs)
    if len(__orfs_to_remove) >= 1:
        if strip_faulty_cds is False:
            err = "{} is an invalid transcript as it has {} invalid ORF{} out of {}. Exceptions: {}".format(
                transcript.id, len(__orfs_to_remove), "" if len(__orfs_to_remove) == 1 else "s",
                __num_orfs, "\n".join(orf_excs))
            transcript.logger.warning(err)
            raise InvalidCDS(err)
        elif len(__orfs_to_remove) == __num_orfs:
            transcript.logger.warning("Every ORF of %s is invalid, stripping the CDS", transcript.id)
            transcript.strip_cds(strand_specific=True)
        else:
            transcript.logger.warning("Stripping %s of %s ORFs out of %s",
                                      transcript.id, len(__orfs_to_remove), __num_orfs)
            for orf_index in reversed(sorted(__orfs_to_remove)):
                internal_orfs.pop(orf_index)
            transcript.internal_orfs = internal_orfs
    else:
        pass

    combined_cds = []
    for orf in transcript.internal_orfs:
        for segment in orf:
            if segment[0] != "CDS":
                continue
            combined_cds.append(segment[1])
    combined_cds = sorted(combined_cds)
    if any([overlap(p1, p2) > 0 for p1, p2 in zip(combined_cds, combined_cds[1:])]):
        internal_orfs = '\n'.join([str(_) for _ in transcript.internal_orfs])
        raise ValueError(f"The internal ORFs of {transcript.id} do overlap each other. This is an error!\n"
                         f"{internal_orfs}\nCombined CDS:{combined_cds}")

    transcript.combined_cds = combined_cds

    if len(transcript.internal_orfs) > 0:
        transcript.selected_internal_orf_index = 0
        _ = transcript.selected_internal_orf


def _fix_stop_codon(transcript):

    """This private function will fix the CDS and stop codons when the transcript comes from GTF2
    and therefore has, incorrectly, the stop codon outside the CDS."""

    if transcript.strand == "-":
        # We need to check whether the stop codon is actually in the same exon.
        if transcript.stop_codon[-1][1] == transcript.combined_cds[0][0] - 1:
            phase = transcript.phases.pop(transcript.combined_cds[0], None)
            transcript.combined_cds[0] = (transcript.stop_codon.pop(-1)[0],
                                      transcript.combined_cds[0][1])
            transcript.phases[transcript.combined_cds[0]] = phase
        transcript.combined_cds = [tuple(_) for _ in transcript.stop_codon] + transcript.combined_cds
        for pos, utr in enumerate(transcript.combined_utr):
            if utr[0] > transcript.combined_cds[-1][1]:
                continue  # Skip the 5'
            over = overlap(utr, transcript.combined_cds[0])
            if over < 0:
                continue
            elif over > 3:
                raise InvalidCDS("Invalid overlap between UTR and CDS found")
            else:
                if over == utr[1] - utr[0] + 1:  # This is equivalent to a fragment. Remove.
                    transcript.combined_utr[pos] = None
                else:
                    transcript.combined_utr[pos] = (utr[0], max(utr[0], transcript.combined_cds[0][0] - 1))
    else:
        # Expand the last CDS
        if transcript.stop_codon[0][0] == transcript.combined_cds[-1][1] + 1:
            phase = transcript.phases.pop(transcript.combined_cds[-1], None)
            transcript.combined_cds[-1] = (transcript.combined_cds[-1][0],
                                           transcript.stop_codon.pop(0)[1])
            transcript.phases[transcript.combined_cds[-1]] = phase
        transcript.combined_cds.extend([tuple(_) for _ in transcript.stop_codon])
        for pos, utr in enumerate(transcript.combined_utr):
            if utr[1] < transcript.combined_cds[0][0]:
                continue  # Skip the 5'
            over = overlap(utr, transcript.combined_cds[-1])
            if over < 0:
                continue
            elif over > 3:
                raise InvalidCDS("Invalid overlap between UTR and CDS found")
            else:
                if over == utr[1] - utr[0] + 1:  # This is equivalent to a fragment. Remove.
                    transcript.combined_utr[pos] = None
                else:
                    transcript.combined_utr[pos] = (min(utr[1], transcript.combined_cds[-1][1] + 1),
                                                    utr[1])
    transcript.combined_utr = [_ for _ in transcript.combined_utr if _ is not None]  # Remove the deleted UTRs
    return transcript


def finalize(transcript):
    """Function to calculate the internal introns from the exons.
    In the first step, it will sort the exons by their internal coordinates.

    :param transcript: the Transcript instance to finalize.
    :type transcript: Mikado.loci.transcript.Transcript

    """

    if transcript.finalized is True:
        return

    transcript.exons = sorted(transcript.exons)

    # Add the stop codon to the CDS
    strip_faulty_cds = getattr(transcript, "strip_faulty_cds", True)

    transcript._cdna_length = None
    _basic_final_checks(transcript)
    # Sort the exons by start then stop

    try:
        _check_cdna_vs_utr(transcript)
    except InvalidCDS as exc:
        if transcript.combined_cds:
            transcript.logger.debug(f"Possible faulty UTR annotation for {transcript.id}, trying to recalculate it.")
            transcript.logger.debug(exc)
            transcript.combined_utr = []
            try:
                _check_cdna_vs_utr(transcript)
            except InvalidCDS as exc:
                if strip_faulty_cds is True:
                    transcript.logger.warning(f"CDS for {transcript.id} completely invalid. Removing it.")
                    transcript.logger.exception(exc)
                    transcript.combined_cds = []
                    transcript.combined_utr = []
                    transcript.internal_orfs = []
                    _basic_final_checks(transcript)
                    _check_cdna_vs_utr(transcript)
                else:
                    raise InvalidCDS(exc)

    transcript.combined_cds = sorted(transcript.combined_cds,
                                     key=operator.itemgetter(0, 1))

    transcript.combined_utr = sorted(transcript.combined_utr,
                                     key=operator.itemgetter(0, 1))

    if transcript.stop_codon:
        transcript.logger.debug("Adding the stop codon to %s", transcript.id)
        transcript.stop_codon = sorted(transcript.stop_codon)
        transcript.combined_cds = sorted(transcript.combined_cds)
        transcript.combined_utr = sorted(transcript.combined_utr)
        stop_in_cds = True
        if transcript.strand == "-" and transcript.combined_cds[0][0] > transcript.stop_codon[-1][1]:
            stop_in_cds = False
        elif transcript.combined_cds[-1][1] < transcript.stop_codon[0][0]:
            stop_in_cds = False
        if not stop_in_cds:
            # Here comes the complicated part
            try:
                transcript = _fix_stop_codon(transcript)
            except InvalidCDS as exc:
                if strip_faulty_cds is True:
                    transcript.strip_cds()
                else:
                    raise InvalidCDS(exc)

    try:
        _check_completeness(transcript)
        transcript.logger.debug("Verifying phase correctness for %s", transcript.id)
        _check_phase_correctness(transcript, strip_faulty_cds=strip_faulty_cds)
        transcript.logger.debug("Calculating intron correctness for %s", transcript.id)
        _calculate_introns(transcript)
    except InvalidCDS:
        if strip_faulty_cds is True:
            transcript.finalized = True
            transcript.logger.debug("Invalid CDS for {}, stripping it", transcript.id)
            transcript.unfinalize()
            return
        else:
            raise

    if transcript.feature == "transcript" and len(transcript.combined_cds) > 0:
        transcript.feature = "mRNA"

    if len(transcript.combined_cds) == 0:
        transcript.selected_internal_orf_cds = tuple([])
    else:
        assert isinstance(transcript.selected_internal_orf_index, int)
        transcript.selected_internal_orf_cds = tuple(
            internal_cds for internal_cds in transcript.internal_orfs[
                transcript.selected_internal_orf_index] if
            internal_cds[0] == "CDS")

    # Create the internal trees
    transcript._calculate_segment_tree()
    transcript._calculate_cds_tree()

    transcript.logger.debug("Checking various attributes")
    if transcript.has_start_codon is None:
        if "has_start_codon" in transcript.attributes:
            transcript.logger.debug("%s has start codon attribute (%s)", transcript.id,
                                    transcript.attributes["has_start_codon"])
            transcript.has_start_codon = transcript.attributes["has_start_codon"]
        else:
            transcript.logger.debug("No predetermined has_start_codon attribute for %s. Attributes: %s",
                                    transcript.id, transcript.attributes)
    else:
        transcript.logger.debug("%s has start codon attribute (%s)", transcript.id, transcript.has_start_codon)

    if transcript.has_stop_codon is None:
        if "has_stop_codon" in transcript.attributes:
            transcript.logger.debug("%s has stop codon attribute (%s)", transcript.id,
                                    transcript.attributes["has_stop_codon"])
            transcript.has_stop_codon = transcript.attributes["has_stop_codon"]
        else:
            transcript.logger.debug("No predetermined has_stop_codon attribute for %s. Attributes: %s",
                                    transcript.id, transcript.attributes)
    else:
        transcript.logger.debug("%s has stop codon attribute (%s)", transcript.id, transcript.has_stop_codon)

    for prop in list(transcript.attributes.keys()):
        val = transcript.attributes[prop]
        if hasattr(transcript, prop):
            try:
                setattr(transcript, prop, val)
            except AttributeError:  # Some instance attributes CANNOT be set from the attributes of the GTF
                transcript.attributes.pop(prop)

    _ = transcript.cdna_length
    transcript._set_basic_lengths()
    transcript._set_distances()
    transcript.finalized = True
    transcript.logger.debug("Finished finalising %s", transcript.id)

    return



import unittest

import networkx

from Mikado._transcripts.clique_methods import find_cliques, find_communities


class TestCliques(unittest.TestCase):

    """"""

    def setUp(self):

        r"""
        We are going to create a graph composed of three disjoint subgraphs:


        0            4                8
          \        /   \            / | \
            2 -- 3  --  5          7 --- 9        12
          /       \   /             \ | /  \
        1           6                10 --  11


        :return:
        """

        self.graph = networkx.Graph()
        self.graph.add_nodes_from([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])

        # First graph
        self.graph.add_edges_from([(0,2), (1,2), (2,3), (3,4),
                                   (3, 5), (3, 6), (4, 5), (5, 6)])
        # Second graph
        self.graph.add_edges_from([(7, 8), (7, 9), (7, 10),
                                   (8, 9), (8, 10), (9, 10),
                                   (9, 11), (10, 11)])
        # Third graph is a singleton

        self.maxDiff = None

        self.correct_cliques = {
            frozenset([0, 2]),
            frozenset([1, 2]),
            frozenset([2, 3]),
            frozenset([3, 4, 5]),
            frozenset([3, 5, 6]),
            frozenset([7, 8, 9, 10]),
            frozenset([9, 10, 11]),
            frozenset([12])}

        self.correct_communities = {
            frozenset([0, 1, 2, 3, 4, 5, 6]),
            frozenset([7, 8, 9, 10, 11]),
            frozenset([12])}

    def test_find_cliques(self):

        cliques = find_cliques(self.graph)
        self.assertEqual(set(cliques),
                         self.correct_cliques,
                         cliques)

    def test_comms(self):
        self.maxDiff = None
        comms = find_communities(self.graph)
        self.assertEqual(comms,
                         self.correct_communities,
                         (comms, self.correct_communities))


if __name__ == '__main__':
    unittest.main()
